# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: messaging/v1/messaging.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'messaging/v1/messaging.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cmessaging/v1/messaging.proto\x12\x15mellions.messaging.v1\x1a\x1fgoogle/protobuf/timestamp.proto\"\x95\x13\n\x0bJobDispatch\x12?\n\x0f\x65tl_job_context\x18\x01 \x01(\x0b\x32$.mellions.messaging.v1.ETLJobContextH\x00\x12M\n\x16\x65tl_completion_context\x18\x03 \x01(\x0b\x32+.mellions.messaging.v1.ETLCompletionContextH\x00\x12U\n\x1aposting_completion_context\x18\x04 \x01(\x0b\x32/.mellions.messaging.v1.PostingCompletionContextH\x00\x12l\n&category_determination_request_context\x18\x06 \x01(\x0b\x32:.mellions.messaging.v1.CategoryDeterminationRequestContextH\x00\x12n\n\'category_determination_response_context\x18\x07 \x01(\x0b\x32;.mellions.messaging.v1.CategoryDeterminationResponseContextH\x00\x12\x45\n\x12\x63oa_update_context\x18\t \x01(\x0b\x32\'.mellions.messaging.v1.COAUpdateContextH\x00\x12O\n\x17webhook_ingress_context\x18\n \x01(\x0b\x32,.mellions.messaging.v1.WebhookIngressContextH\x00\x12s\n*posting_rule_determination_request_context\x18\x0b \x01(\x0b\x32=.mellions.messaging.v1.PostingRuleDeterminationRequestContextH\x00\x12K\n\x15generic_event_context\x18\x0c \x01(\x0b\x32*.mellions.messaging.v1.GenericEventContextH\x00\x12\x45\n\x12goal_event_context\x18\r \x01(\x0b\x32\'.mellions.messaging.v1.GoalEventContextH\x00\x12\x45\n\x12\x64\x61ta_ready_context\x18\x0e \x01(\x0b\x32\'.mellions.messaging.v1.DataReadyContextH\x00\x12M\n\x16goal_operation_request\x18\x0f \x01(\x0b\x32+.mellions.messaging.v1.GoalOperationRequestH\x00\x12O\n\x17goal_operation_response\x18\x10 \x01(\x0b\x32,.mellions.messaging.v1.GoalOperationResponseH\x00\x12I\n\x14goal_subledger_event\x18\x11 \x01(\x0b\x32).mellions.messaging.v1.GoalSubledgerEventH\x00\x12\x42\n\x10suggestion_event\x18\x12 \x01(\x0b\x32&.mellions.messaging.v1.SuggestionEventH\x00\x12u\n+posting_rule_determination_response_context\x18\x13 \x01(\x0b\x32>.mellions.messaging.v1.PostingRuleDeterminationResponseContextH\x00\x12M\n\x16\x64\x61ta_available_context\x18\x14 \x01(\x0b\x32+.mellions.messaging.v1.DataAvailableContextH\x00\x12O\n\x17insight_trigger_context\x18\x16 \x01(\x0b\x32,.mellions.messaging.v1.InsightTriggerContextH\x00\x12Q\n\x18\x62\x61\x63kfill_trigger_context\x18\x17 \x01(\x0b\x32-.mellions.messaging.v1.BackfillTriggerContextH\x00\x12\x45\n\x12\x65tl_failed_context\x18\x1b \x01(\x0b\x32\'.mellions.messaging.v1.ETLFailedContextH\x00\x12M\n\x16posting_failed_context\x18\x1c \x01(\x0b\x32+.mellions.messaging.v1.PostingFailedContextH\x00\x12O\n\x18\x66_x_rate_request_context\x18\x1d \x01(\x0b\x32+.mellions.messaging.v1.FXRateRequestContextH\x00\x12Q\n\x19\x66_x_rate_response_context\x18\x1e \x01(\x0b\x32,.mellions.messaging.v1.FXRateResponseContextH\x00\x12`\n insight_analysis_request_context\x18! \x01(\x0b\x32\x34.mellions.messaging.v1.InsightAnalysisRequestContextH\x00\x12\x62\n!insight_analysis_response_context\x18\" \x01(\x0b\x32\x35.mellions.messaging.v1.InsightAnalysisResponseContextH\x00\x12Q\n\x18taxonomy_release_context\x18# \x01(\x0b\x32-.mellions.messaging.v1.TaxonomyReleaseContextH\x00\x12Z\n\x1dposting_rules_release_context\x18$ \x01(\x0b\x32\x31.mellions.messaging.v1.PostingRulesReleaseContextH\x00\x12[\n\x1d\x63onnection_reconciled_context\x18% \x01(\x0b\x32\x32.mellions.messaging.v1.ConnectionReconciledContextH\x00\x42\n\n\x08job_typeJ\x04\x08\x02\x10\x03J\x04\x08\x05\x10\x06J\x04\x08\x08\x10\tJ\x04\x08\x15\x10\x16J\x04\x08\x18\x10\x19J\x04\x08\x1a\x10\x1bJ\x04\x08\x1f\x10 \"\xe2\x01\n\rETLJobContext\x12\x32\n\tsync_type\x18\x02 \x01(\x0e\x32\x1f.mellions.messaging.v1.SyncType\x12.\n\nstart_date\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_date\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05\x63ount\x18\x05 \x01(\x05\x12\x30\n\x04\x64\x61ta\x18\x06 \x03(\x0b\x32\".mellions.messaging.v1.DataDetails\"\xb4\x01\n\x14\x45TLCompletionContext\x12\x44\n\x12processing_summary\x18\x02 \x01(\x0b\x32(.mellions.messaging.v1.ProcessingSummary\x12\x42\n\x0fquality_metrics\x18\x03 \x01(\x0b\x32).mellions.messaging.v1.DataQualityMetrics\x12\x12\n\nnext_steps\x18\x04 \x03(\t\"\xc5\x01\n\x18PostingCompletionContext\x12\x0e\n\x06job_id\x18\x02 \x01(\t\x12>\n\x0fposting_summary\x18\x03 \x01(\x0b\x32%.mellions.messaging.v1.PostingSummary\x12\x17\n\x0fjournal_entries\x18\x04 \x03(\t\x12\x19\n\x11\x61i_required_items\x18\x05 \x03(\t\x12\x12\n\nnext_steps\x18\x06 \x03(\t\x12\x11\n\tfinalized\x18\x07 \x01(\x08\"\xdc\x03\n\x16\x42\x61\x63kfillTriggerContext\x12\x15\n\rconnection_id\x18\x02 \x01(\t\x12\x10\n\x08provider\x18\x03 \x01(\t\x12\x33\n\ndata_types\x18\x04 \x03(\x0e\x32\x1f.mellions.messaging.v1.DataType\x12\x30\n\x0ctriggered_at\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x16\n\x0etrigger_source\x18\x06 \x01(\t\x12\x33\n\nstart_date\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x00\x88\x01\x01\x12\x31\n\x08\x65nd_date\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x01\x88\x01\x01\x12\x16\n\x0e\x63orrelation_id\x18\t \x01(\t\x12M\n\x08metadata\x18\n \x03(\x0b\x32;.mellions.messaging.v1.BackfillTriggerContext.MetadataEntry\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\r\n\x0b_start_dateB\x0b\n\t_end_date\"\xaf\x02\n\x15WebhookIngressContext\x12\x10\n\x08provider\x18\x02 \x01(\t\x12\x12\n\nevent_type\x18\x03 \x01(\t\x12\x15\n\rconnection_id\x18\x04 \x01(\t\x12\x13\n\x0braw_payload\x18\x05 \x01(\x0c\x12\x17\n\x0fidempotency_key\x18\x06 \x01(\t\x12/\n\x0breceived_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12J\n\x07headers\x18\x08 \x03(\x0b\x32\x39.mellions.messaging.v1.WebhookIngressContext.HeadersEntry\x1a.\n\x0cHeadersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xc5\x02\n\x14GoalOperationRequest\x12\x12\n\nrequest_id\x18\x02 \x01(\t\x12\x0f\n\x07goal_id\x18\x03 \x01(\t\x12\x17\n\x0fidempotency_key\x18\x04 \x01(\t\x12\x38\n\x07\x65\x61rmark\x18\n \x01(\x0b\x32%.mellions.messaging.v1.EarmarkRequestH\x00\x12\x34\n\x05spend\x18\x0b \x01(\x0b\x32#.mellions.messaging.v1.SpendRequestH\x00\x12:\n\x08\x61llocate\x18\x0c \x01(\x0b\x32&.mellions.messaging.v1.AllocateRequestH\x00\x12\x36\n\x06refund\x18\r \x01(\x0b\x32$.mellions.messaging.v1.RefundRequestH\x00\x42\x0b\n\toperation\"\xf7\x02\n\x15GoalOperationResponse\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12\x0f\n\x07success\x18\x02 \x01(\x08\x12\x12\n\nerror_code\x18\x03 \x01(\t\x12\x15\n\rerror_message\x18\x04 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x05 \x01(\t\x12\x36\n\x07\x65ntries\x18\n \x03(\x0b\x32%.mellions.messaging.v1.SubledgerEntry\x12^\n\x12\x61ggregates_by_goal\x18\x0b \x03(\x0b\x32\x42.mellions.messaging.v1.GoalOperationResponse.AggregatesByGoalEntry\x1a^\n\x15\x41ggregatesByGoalEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x34\n\x05value\x18\x02 \x01(\x0b\x32%.mellions.messaging.v1.GoalAggregates:\x02\x38\x01\"\xd3\x01\n\x12GoalSubledgerEvent\x12\x10\n\x08\x65vent_id\x18\x02 \x01(\t\x12\x0f\n\x07goal_id\x18\x03 \x01(\t\x12\x11\n\toperation\x18\x04 \x01(\t\x12\x34\n\x05\x65ntry\x18\x05 \x01(\x0b\x32%.mellions.messaging.v1.SubledgerEntry\x12\x39\n\naggregates\x18\x06 \x01(\x0b\x32%.mellions.messaging.v1.GoalAggregates\x12\x16\n\x0e\x63orrelation_id\x18\x07 \x01(\t\"\xaf\x02\n\x15InsightTriggerContext\x12\x14\n\x0cperiod_month\x18\x02 \x01(\t\x12\x18\n\x10max_entry_cursor\x18\x03 \x01(\t\x12\x10\n\x08\x62\x61tch_id\x18\x04 \x01(\t\x12\x15\n\rentries_count\x18\x05 \x01(\x05\x12/\n\x0boccurred_at\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x16\n\x0e\x63orrelation_id\x18\x07 \x01(\t\x12\x46\n\x05hints\x18\x08 \x03(\x0b\x32\x37.mellions.messaging.v1.InsightTriggerContext.HintsEntry\x1a,\n\nHintsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xe2\x01\n\x10\x44\x61taReadyContext\x12\x0e\n\x06job_id\x18\x02 \x01(\t\x12\x1e\n\x16latest_posting_version\x18\x03 \x01(\t\x12\x1a\n\x12ledger_snapshot_id\x18\x04 \x01(\t\x12\x1b\n\x13transactions_posted\x18\x05 \x01(\x05\x12\x1a\n\x12transactions_total\x18\x06 \x01(\x05\x12\x1b\n\x13unreviewed_ai_count\x18\x07 \x01(\x05\x12\x19\n\x11\x61i_required_items\x18\x08 \x03(\t\x12\x11\n\tfinalized\x18\t \x01(\x08\"L\n#CategoryDeterminationRequestContext\x12\x12\n\nrequest_id\x18\x02 \x01(\t\x12\x11\n\ttxn_count\x18\x03 \x01(\x05\"\xbd\x01\n$CategoryDeterminationResponseContext\x12\x12\n\nrequest_id\x18\x02 \x01(\t\x12\x37\n\x06status\x18\x03 \x01(\x0e\x32\'.mellions.messaging.v1.ProcessingStatus\x12\x16\n\x0e\x65nriched_count\x18\x04 \x01(\x05\x12\x14\n\x0c\x66\x61iled_count\x18\x05 \x01(\x05\x12\x1a\n\x12processing_time_ms\x18\x06 \x01(\x03\"\x95\x01\n&PostingRuleDeterminationRequestContext\x12\x12\n\nrequest_id\x18\x02 \x01(\t\x12\x11\n\ttxn_count\x18\x03 \x01(\x05\x12\x44\n\x12\x64\x65termination_type\x18\x04 \x01(\x0e\x32(.mellions.messaging.v1.DeterminationType\"\xc0\x01\n\'PostingRuleDeterminationResponseContext\x12\x12\n\nrequest_id\x18\x02 \x01(\t\x12\x37\n\x06status\x18\x03 \x01(\x0e\x32\'.mellions.messaging.v1.ProcessingStatus\x12\x16\n\x0eresolved_count\x18\x04 \x01(\x05\x12\x14\n\x0c\x66\x61iled_count\x18\x05 \x01(\x05\x12\x1a\n\x12processing_time_ms\x18\x06 \x01(\x03\"\xcf\x01\n\x1dInsightAnalysisRequestContext\x12\x12\n\nrequest_id\x18\x02 \x01(\t\x12\x38\n\x0cinsight_type\x18\x03 \x01(\x0e\x32\".mellions.messaging.v1.InsightType\x12\x30\n\x0cperiod_start\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nperiod_end\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xc3\x01\n\x1eInsightAnalysisResponseContext\x12\x12\n\nrequest_id\x18\x02 \x01(\t\x12\x37\n\x06status\x18\x03 \x01(\x0e\x32\'.mellions.messaging.v1.ProcessingStatus\x12\x38\n\x0cinsight_type\x18\x04 \x01(\x0e\x32\".mellions.messaging.v1.InsightType\x12\x1a\n\x12processing_time_ms\x18\x05 \x01(\x03\"\x83\x02\n\x16TaxonomyReleaseContext\x12\x12\n\nrelease_id\x18\x02 \x01(\t\x12:\n\rtaxonomy_type\x18\x03 \x01(\x0e\x32#.mellions.messaging.v1.TaxonomyType\x12\x0f\n\x07version\x18\x04 \x01(\x05\x12\x15\n\rsha256_digest\x18\x05 \x01(\t\x12\x30\n\x0crelease_date\x18\x32 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12?\n\x0erelease_status\x18\x06 \x01(\x0e\x32\'.mellions.messaging.v1.COAReleaseStatus\"\xdf\x01\n\x1aPostingRulesReleaseContext\x12\x12\n\nrelease_id\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\x05\x12\x15\n\rsha256_digest\x18\x04 \x01(\t\x12\x12\n\nrule_count\x18\x05 \x01(\x05\x12\x30\n\x0crelease_date\x18\x32 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12?\n\x0erelease_status\x18\x06 \x01(\x0e\x32\'.mellions.messaging.v1.COAReleaseStatus\"\xd5\x01\n\x10\x43OAUpdateContext\x12\x17\n\x0frelease_version\x18\x02 \x01(\t\x12\x30\n\x0crelease_date\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12?\n\x0erelease_status\x18\x04 \x01(\x0e\x32\'.mellions.messaging.v1.COAReleaseStatus\x12\x35\n\x04type\x18\x05 \x01(\x0e\x32\'.mellions.messaging.v1.COAApplicability\"\xad\x02\n\x13GenericEventContext\x12\x10\n\x08\x63\x61tegory\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12N\n\nattributes\x18\x04 \x03(\x0b\x32:.mellions.messaging.v1.GenericEventContext.AttributesEntry\x12\x13\n\x0braw_payload\x18\x05 \x01(\x0c\x12\x14\n\x0c\x63ontent_type\x18\x06 \x01(\t\x12\x17\n\x0fidempotency_key\x18\x07 \x01(\t\x12/\n\x0boccurred_at\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a\x31\n\x0f\x41ttributesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xdd\x01\n\x10GoalEventContext\x12\x0f\n\x07goal_id\x18\x02 \x01(\t\x12@\n\x0eoperation_type\x18\x03 \x01(\x0e\x32(.mellions.messaging.v1.GoalOperationType\x12\x38\n\x07payload\x18\x04 \x01(\x0b\x32\'.mellions.messaging.v1.GoalEventPayload\x12<\n\nallocation\x18\x05 \x01(\x0b\x32(.mellions.messaging.v1.AllocationContext\"\xc0\x02\n\x0fSuggestionEvent\x12\x10\n\x08\x65vent_id\x18\x02 \x01(\t\x12\x16\n\x0etransaction_id\x18\x03 \x01(\t\x12\x15\n\rsuggestion_id\x18\x04 \x01(\t\x12\x12\n\nconfidence\x18\x05 \x01(\x01\x12>\n\ncandidates\x18\x06 \x03(\x0b\x32*.mellions.messaging.v1.SuggestionCandidate\x12\x37\n\x06status\x18\x07 \x01(\x0e\x32\'.mellions.messaging.v1.SuggestionStatus\x12.\n\ncreated_at\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bresolved_at\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x88\x02\n\x14\x44\x61taAvailableContext\x12\x10\n\x08\x62\x61tch_id\x18\x02 \x01(\t\x12\x12\n\nitem_count\x18\x03 \x01(\x05\x12\x36\n\x0b\x64\x61ta_status\x18\x04 \x01(\x0e\x32!.mellions.messaging.v1.DataStatus\x12:\n\rposting_state\x18\x05 \x01(\x0e\x32#.mellions.messaging.v1.PostingState\x12\x16\n\x0e\x66\x61ilure_reason\x18\x06 \x01(\t\x12\x14\n\x0cposted_count\x18\x07 \x01(\x05\x12\x14\n\x0c\x66\x61iled_count\x18\x08 \x01(\x05\x12\x12\n\nfailed_ids\x18\t \x03(\t\"\xc9\x01\n\x10\x45TLFailedContext\x12\x0e\n\x06job_id\x18\x02 \x01(\t\x12\x12\n\nerror_code\x18\x03 \x01(\t\x12\x15\n\rerror_message\x18\x04 \x01(\t\x12\x17\n\x0f\x66\x61iled_at_stage\x18\x05 \x01(\t\x12\x19\n\x11records_processed\x18\x06 \x01(\x05\x12\x16\n\x0erecords_failed\x18\x07 \x01(\x05\x12\x11\n\tretriable\x18\x08 \x01(\x08\x12\x1b\n\x13original_message_id\x18\t \x01(\t\"\xb8\x01\n\x14PostingFailedContext\x12\x0e\n\x06job_id\x18\x02 \x01(\t\x12\x12\n\nerror_code\x18\x03 \x01(\t\x12\x15\n\rerror_message\x18\x04 \x01(\t\x12\x1e\n\x16\x66\x61iled_transaction_ids\x18\x05 \x03(\t\x12\x15\n\rsuccess_count\x18\x06 \x01(\x05\x12\x11\n\tretriable\x18\x07 \x01(\x08\x12\x1b\n\x13original_message_id\x18\x08 \x01(\t\"\xb0\x02\n\x14\x46XRateRequestContext\x12>\n\x0crequest_type\x18\x02 \x01(\x0e\x32(.mellions.messaging.v1.FXRateRequestType\x12\x16\n\x0e\x63urrency_codes\x18\x03 \x03(\t\x12\x15\n\rbase_currency\x18\x04 \x01(\t\x12.\n\nas_of_date\x18\x32 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nstart_date\x18\x33 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_date\x18\x34 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1b\n\x13is_fetch_if_missing\x18\t \x01(\x08\"\xd1\x01\n\x15\x46XRateResponseContext\x12\x15\n\rbase_currency\x18\x02 \x01(\t\x12\x31\n\x05rates\x18\x03 \x03(\x0b\x32\".mellions.messaging.v1.FXRateEntry\x12;\n\x06status\x18\x05 \x01(\x0e\x32+.mellions.messaging.v1.FXRateResponseStatus\x12\x15\n\rerror_message\x18\x06 \x01(\t\x12\x1a\n\x12missing_currencies\x18\x07 \x03(\t\"\xfe\x01\n\x1b\x43onnectionReconciledContext\x12\x15\n\rconnection_id\x18\x01 \x01(\t\x12!\n\x19merged_into_connection_id\x18\x02 \x01(\t\x12\x0f\n\x07user_id\x18\x03 \x01(\t\x12\x11\n\tentity_id\x18\x04 \x01(\t\x12\x10\n\x08provider\x18\x05 \x01(\t\x12G\n\x12\x64uplicate_accounts\x18\x06 \x03(\x0b\x32+.mellions.messaging.v1.DuplicateAccountPair\x12\x12\n\nmatch_tier\x18\x07 \x01(\t\x12\x12\n\nmatch_hash\x18\x08 \x01(\t\"L\n\x14\x44uplicateAccountPair\x12\x19\n\x11source_account_id\x18\x01 \x01(\t\x12\x19\n\x11target_account_id\x18\x02 \x01(\t\"\xa2\x02\n\x0fIdentityContext\x12\x0f\n\x07user_id\x18\x01 \x01(\t\x12\x11\n\tentity_id\x18\x02 \x01(\t\x12\x0f\n\x07item_id\x18\x03 \x01(\t\x12\x10\n\x08\x62\x61tch_id\x18\x04 \x01(\t\x12\x32\n\tdata_type\x18\x05 \x01(\x0e\x32\x1f.mellions.messaging.v1.DataType\x12\x0e\n\x06job_id\x18\n \x01(\t\x12\x12\n\nrequest_id\x18\x0b \x01(\t\x12\x10\n\x08\x65vent_id\x18\x0c \x01(\t\x12\x15\n\rconnection_id\x18\x14 \x01(\t\x12\x10\n\x08provider\x18\x15 \x01(\t\x12\x16\n\x0etrigger_source\x18\x16 \x01(\t\x12\x17\n\x0fidempotency_key\x18\x1e \x01(\tJ\x04\x08\x06\x10\x07\"\xdb\x02\n\rMessageHeader\x12\x12\n\nmessage_id\x18\x01 \x01(\t\x12\x16\n\x0e\x63orrelation_id\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\x12<\n\x0esource_service\x18\x04 \x01(\x0e\x32$.mellions.messaging.v1.SourceService\x12<\n\x0etarget_service\x18\x05 \x01(\x0e\x32$.mellions.messaging.v1.TargetService\x12\x38\n\x0cmessage_type\x18\x06 \x01(\x0e\x32\".mellions.messaging.v1.MessageType\x12\x0f\n\x07version\x18\x07 \x01(\t\x12\x38\n\x08identity\x18\n \x01(\x0b\x32&.mellions.messaging.v1.IdentityContextJ\x04\x08\x08\x10\tJ\x04\x08\t\x10\n\"\xe2\x01\n\x11ProcessingSummary\x12\x19\n\x11records_processed\x18\x01 \x01(\x05\x12\x16\n\x0erecords_failed\x18\x02 \x01(\x05\x12\x17\n\x0frecords_skipped\x18\x03 \x01(\x05\x12\x1a\n\x12processing_time_ms\x18\x04 \x01(\x03\x12\x37\n\x06status\x18\x05 \x01(\x0e\x32\'.mellions.messaging.v1.ProcessingStatus\x12\x15\n\rquality_score\x18\x06 \x01(\x01\x12\x15\n\rerror_summary\x18\x07 \x03(\t\"\xad\x03\n\x12\x44\x61taQualityMetrics\x12\x15\n\roverall_score\x18\x01 \x01(\x01\x12\\\n\x12\x66ield_completeness\x18\x02 \x03(\x0b\x32@.mellions.messaging.v1.DataQualityMetrics.FieldCompletenessEntry\x12\\\n\x12validation_results\x18\x03 \x03(\x0b\x32@.mellions.messaging.v1.DataQualityMetrics.ValidationResultsEntry\x12\x19\n\x11\x63\x61tegory_accuracy\x18\x04 \x01(\x01\x12\x18\n\x10\x64uplication_rate\x18\x05 \x01(\x01\x12\x1b\n\x13\x65nrichment_coverage\x18\x06 \x01(\x01\x1a\x38\n\x16\x46ieldCompletenessEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x01:\x02\x38\x01\x1a\x38\n\x16ValidationResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\"\x9a\x02\n\x0ePostingSummary\x12\x1a\n\x12transactions_total\x18\x01 \x01(\x05\x12\x1b\n\x13transactions_posted\x18\x02 \x01(\x05\x12\x1b\n\x13transactions_failed\x18\x03 \x01(\x05\x12 \n\x18transactions_ai_required\x18\x04 \x01(\x05\x12\x1f\n\x17journal_entries_created\x18\x05 \x01(\x05\x12\x1a\n\x12processing_time_ms\x18\x06 \x01(\x03\x12\x37\n\x06status\x18\x07 \x01(\x0e\x32\'.mellions.messaging.v1.ProcessingStatus\x12\x1a\n\x12\x62\x61lance_validation\x18\x08 \x01(\x08\"3\n\x0b\x44\x61taDetails\x12\x12\n\naccount_id\x18\x01 \x01(\t\x12\x10\n\x08\x64\x61ta_ids\x18\x02 \x03(\t\"\x93\x01\n\nPostingLeg\x12\x14\n\x0c\x61\x63\x63ount_code\x18\x01 \x01(\t\x12\x14\n\x0c\x61\x63\x63ount_name\x18\x02 \x01(\t\x12\x34\n\nentry_side\x18\x03 \x01(\x0e\x32 .mellions.messaging.v1.EntrySide\x12\x0e\n\x06\x61mount\x18\x04 \x01(\x01\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\"\xb8\x03\n\x10GoalEventPayload\x12\r\n\x05title\x18\x01 \x01(\t\x12\x14\n\x0c\x63\x61tegory_key\x18\x02 \x01(\t\x12\x15\n\rtarget_amount\x18\x03 \x01(\x01\x12\x17\n\x0f\x63urrent_balance\x18\x04 \x01(\x01\x12\x10\n\x08\x63urrency\x18\x05 \x01(\t\x12.\n\nstart_date\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0btarget_date\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x31\n\x06status\x18\x08 \x01(\x0e\x32!.mellions.messaging.v1.GoalStatus\x12\x10\n\x08priority\x18\t \x01(\x05\x12;\n\x0bperformance\x18\n \x01(\x0e\x32&.mellions.messaging.v1.GoalPerformance\x12\x46\n\x13\x61llocation_settings\x18\x0b \x01(\x0b\x32).mellions.messaging.v1.AllocationSettings\x12\x12\n\naccount_id\x18\x0c \x01(\t\"\xaf\x02\n\x12\x41llocationSettings\x12>\n\x0f\x61llocation_type\x18\x01 \x01(\x0e\x32%.mellions.messaging.v1.AllocationType\x12\x19\n\x11\x61llocation_amount\x18\x02 \x01(\x01\x12\x1d\n\x15\x61llocation_percentage\x18\x03 \x01(\x01\x12=\n\tfrequency\x18\x04 \x01(\x0e\x32*.mellions.messaging.v1.AllocationFrequency\x12\x16\n\x0e\x61llocation_day\x18\x05 \x01(\x05\x12\x11\n\tis_active\x18\x06 \x01(\x08\x12\x35\n\x11next_trigger_date\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x81\x02\n\x11\x41llocationContext\x12\x15\n\rallocation_id\x18\x01 \x01(\t\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x01\x12\x19\n\x11source_account_id\x18\x03 \x01(\t\x12\x1e\n\x16\x64\x65stination_account_id\x18\x04 \x01(\t\x12\x0e\n\x06status\x18\x05 \x01(\t\x12\x16\n\x0e\x66\x61ilure_reason\x18\x06 \x01(\t\x12\x30\n\x0ctriggered_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x30\n\x0c\x63ompleted_at\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xd3\x01\n\x0e\x45\x61rmarkRequest\x12\x0e\n\x06\x61mount\x18\x01 \x01(\x01\x12\x10\n\x08\x63urrency\x18\x02 \x01(\t\x12\x14\n\x0cposting_date\x18\x03 \x01(\t\x12R\n\x0fsource_metadata\x18\x04 \x03(\x0b\x32\x39.mellions.messaging.v1.EarmarkRequest.SourceMetadataEntry\x1a\x35\n\x13SourceMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"l\n\x0cSpendRequest\x12\x0e\n\x06\x61mount\x18\x01 \x01(\x01\x12\x10\n\x08\x63urrency\x18\x02 \x01(\t\x12\x16\n\x0etransaction_id\x18\x03 \x01(\t\x12\x14\n\x0cposting_date\x18\x04 \x01(\t\x12\x0c\n\x04note\x18\x05 \x01(\t\"\x89\x01\n\x0f\x41llocateRequest\x12\x16\n\x0etransaction_id\x18\x01 \x01(\t\x12\x10\n\x08\x63urrency\x18\x02 \x01(\t\x12\x14\n\x0cposting_date\x18\x03 \x01(\t\x12\x36\n\x0b\x61llocations\x18\x04 \x03(\x0b\x32!.mellions.messaging.v1.Allocation\";\n\nAllocation\x12\x0f\n\x07goal_id\x18\x01 \x01(\t\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x01\x12\x0c\n\x04note\x18\x03 \x01(\t\"\x88\x01\n\rRefundRequest\x12\x0e\n\x06\x61mount\x18\x01 \x01(\x01\x12\x10\n\x08\x63urrency\x18\x02 \x01(\t\x12\x19\n\x11original_entry_id\x18\x03 \x01(\t\x12\x16\n\x0etransaction_id\x18\x04 \x01(\t\x12\x14\n\x0cposting_date\x18\x05 \x01(\t\x12\x0c\n\x04note\x18\x06 \x01(\t\"\x9d\x02\n\x0eSubledgerEntry\x12\x10\n\x08\x65ntry_id\x18\x01 \x01(\t\x12\x0f\n\x07goal_id\x18\x02 \x01(\t\x12\x38\n\nentry_type\x18\x03 \x01(\x0e\x32$.mellions.messaging.v1.GoalEntryType\x12\x0e\n\x06\x61mount\x18\x04 \x01(\x01\x12\x10\n\x08\x63urrency\x18\x05 \x01(\t\x12\x14\n\x0cposting_date\x18\x06 \x01(\t\x12.\n\ncreated_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x16\n\x0etransaction_id\x18\x08 \x01(\t\x12\x13\n\x0bis_reversed\x18\t \x01(\x08\x12\x19\n\x11original_entry_id\x18\n \x01(\t\"\xdd\x01\n\x0eGoalAggregates\x12\x17\n\x0ftotal_allocated\x18\x01 \x01(\x01\x12\x13\n\x0btotal_spent\x18\x02 \x01(\x01\x12\x16\n\x0etotal_refunded\x18\x03 \x01(\x01\x12\x11\n\tnet_spent\x18\x04 \x01(\x01\x12\x11\n\tremaining\x18\x05 \x01(\x01\x12\x18\n\x10\x66unding_progress\x18\x06 \x01(\x01\x12\x1c\n\x14\x63onsumption_progress\x18\x07 \x01(\x01\x12\x15\n\rtarget_amount\x18\x08 \x01(\x01\x12\x10\n\x08\x63urrency\x18\t \x01(\t\"o\n\x13SuggestionCandidate\x12\x0f\n\x07goal_id\x18\x01 \x01(\t\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x01\x12\x10\n\x08\x63urrency\x18\x03 \x01(\t\x12\x12\n\nconfidence\x18\x04 \x01(\x01\x12\x11\n\trationale\x18\x05 \x01(\t\"\x98\x01\n\x0b\x46XRateEntry\x12\x15\n\rcurrency_code\x18\x01 \x01(\t\x12\x0c\n\x04rate\x18\x02 \x01(\t\x12\x32\n\x0e\x65\x66\x66\x65\x63tive_date\x18\x32 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06source\x18\x04 \x01(\t\x12 \n\x18is_fetched_from_provider\x18\x05 \x01(\x08*\x96\x02\n\x10ProcessingStatus\x12!\n\x1dPROCESSING_STATUS_UNSPECIFIED\x10\x00\x12\x1d\n\x19PROCESSING_STATUS_PENDING\x10\x01\x12 \n\x1cPROCESSING_STATUS_PROCESSING\x10\x02\x12\x1d\n\x19PROCESSING_STATUS_SUCCESS\x10\x03\x12\x1c\n\x18PROCESSING_STATUS_FAILED\x10\x04\x12\x1b\n\x17PROCESSING_STATUS_RETRY\x10\x05\x12\x1d\n\x19PROCESSING_STATUS_TIMEOUT\x10\x06\x12%\n!PROCESSING_STATUS_PARTIAL_SUCCESS\x10\x07*\x90\x03\n\x08\x44\x61taType\x12\x19\n\x15\x44\x41TA_TYPE_UNSPECIFIED\x10\x00\x12\x1a\n\x16\x44\x41TA_TYPE_TRANSACTIONS\x10\x01\x12\x16\n\x12\x44\x41TA_TYPE_ACCOUNTS\x10\x02\x12\x19\n\x15\x44\x41TA_TYPE_LIABILITIES\x10\x03\x12\x19\n\x15\x44\x41TA_TYPE_INVESTMENTS\x10\x04\x12\x16\n\x12\x44\x41TA_TYPE_IDENTITY\x10\x05\x12\x14\n\x10\x44\x41TA_TYPE_ASSETS\x10\x06\x12\x14\n\x10\x44\x41TA_TYPE_INCOME\x10\x07\x12\x1c\n\x18\x44\x41TA_TYPE_CREDIT_DETAILS\x10\x08\x12\x19\n\x15\x44\x41TA_TYPE_CONNECTIONS\x10\t\x12\x18\n\x14\x44\x41TA_TYPE_STATEMENTS\x10\n\x12\x13\n\x0f\x44\x41TA_TYPE_GOALS\x10\x0b\x12\x1f\n\x1b\x44\x41TA_TYPE_DASHBOARD_WIDGETS\x10\x0c\x12\x19\n\x15\x44\x41TA_TYPE_INSTITUTION\x10\r\x12\x17\n\x13\x44\x41TA_TYPE_RECURRING\x10\x0e*\x96\x06\n\rSourceService\x12\x1e\n\x1aSOURCE_SERVICE_UNSPECIFIED\x10\x00\x12\x1b\n\x17SOURCE_SERVICE_NEONLINK\x10\x01\x12\x1e\n\x1aSOURCE_SERVICE_DATASERVICE\x10\x02\x12\"\n\x1eSOURCE_SERVICE_POSTING_SERVICE\x10\x03\x12 \n\x1cSOURCE_SERVICE_FINAI_SERVICE\x10\x04\x12\x1f\n\x1bSOURCE_SERVICE_SYNC_SERVICE\x10\x05\x12\"\n\x1eSOURCE_SERVICE_WEBHOOK_SERVICE\x10\x06\x12\x1f\n\x1bSOURCE_SERVICE_GOAL_SERVICE\x10\x07\x12\"\n\x1eSOURCE_SERVICE_FINAPIS_SERVICE\x10\x08\x12$\n SOURCE_SERVICE_DASHBOARD_SERVICE\x10\t\x12\x1c\n\x18SOURCE_SERVICE_CONNECTOR\x10\n\x12!\n\x1dSOURCE_SERVICE_FXRATE_SERVICE\x10\x0b\x12\x1c\n\x18SOURCE_SERVICE_MOBILEAPP\x10\x14\x12\x19\n\x15SOURCE_SERVICE_WEBAPP\x10\x15\x12\x18\n\x14SOURCE_SERVICE_PLAID\x10\x1e\x12\x19\n\x15SOURCE_SERVICE_STRIPE\x10\x1f\x12\x19\n\x15SOURCE_SERVICE_PAYPAL\x10 \x12\x1e\n\x1aSOURCE_SERVICE_MASTER_CARD\x10!\x12\x17\n\x13SOURCE_SERVICE_VISA\x10\"\x12\x1d\n\x19SOURCE_SERVICE_QUICKBOOKS\x10#\x12 \n\x1cSOURCE_SERVICE_MANUAL_UPLOAD\x10$\x12\x17\n\x13SOURCE_SERVICE_XERO\x10%\x12\x1b\n\x17SOURCE_SERVICE_BANK_API\x10&\x12\x1d\n\x19SOURCE_SERVICE_CSV_IMPORT\x10\'\x12\x19\n\x15SOURCE_SERVICE_FLINKS\x10(*\xad\x03\n\rTargetService\x12\x1e\n\x1aTARGET_SERVICE_UNSPECIFIED\x10\x00\x12\x1e\n\x1aTARGET_SERVICE_ETL_SERVICE\x10\x01\x12\"\n\x1eTARGET_SERVICE_POSTING_SERVICE\x10\x02\x12 \n\x1cTARGET_SERVICE_FINAI_SERVICE\x10\x03\x12\x1f\n\x1bTARGET_SERVICE_SYNC_SERVICE\x10\x04\x12\"\n\x1eTARGET_SERVICE_WEBHOOK_SERVICE\x10\x05\x12%\n!TARGET_SERVICE_MONITORING_SERVICE\x10\x06\x12\x1f\n\x1bTARGET_SERVICE_GOAL_SERVICE\x10\x07\x12\"\n\x1eTARGET_SERVICE_FINAPIS_SERVICE\x10\x08\x12$\n TARGET_SERVICE_DASHBOARD_SERVICE\x10\t\x12\x1c\n\x18TARGET_SERVICE_CONNECTOR\x10\n\x12!\n\x1dTARGET_SERVICE_FXRATE_SERVICE\x10\x0b*\xb0\x01\n\x08SyncType\x12\x19\n\x15SYNC_TYPE_UNSPECIFIED\x10\x00\x12\x19\n\x15SYNC_TYPE_INCREMENTAL\x10\x01\x12\x12\n\x0eSYNC_TYPE_FULL\x10\x02\x12\x13\n\x0fSYNC_TYPE_DELTA\x10\x03\x12\x15\n\x11SYNC_TYPE_INITIAL\x10\x04\x12\x14\n\x10SYNC_TYPE_DELETE\x10\x05\x12\x18\n\x14SYNC_TYPE_HISTORICAL\x10\x06*T\n\tEntrySide\x12\x1a\n\x16\x45NTRY_SIDE_UNSPECIFIED\x10\x00\x12\x14\n\x10\x45NTRY_SIDE_DEBIT\x10\x01\x12\x15\n\x11\x45NTRY_SIDE_CREDIT\x10\x02*\xa3\x01\n\x11\x44\x65terminationType\x12\"\n\x1e\x44\x45TERMINATION_TYPE_UNSPECIFIED\x10\x00\x12\x1e\n\x1a\x44\x45TERMINATION_TYPE_INITIAL\x10\x01\x12\'\n#DETERMINATION_TYPE_RECLASSIFICATION\x10\x02\x12!\n\x1d\x44\x45TERMINATION_TYPE_CORRECTION\x10\x03*\xb9\x01\n\x0bInsightType\x12\x1c\n\x18INSIGHT_TYPE_UNSPECIFIED\x10\x00\x12\x1e\n\x1aINSIGHT_TYPE_SAFE_TO_SPEND\x10\x01\x12\'\n#INSIGHT_TYPE_FINANCIAL_HEALTH_SCORE\x10\x02\x12\x1f\n\x1bINSIGHT_TYPE_SPENDING_TREND\x10\x03\x12\"\n\x1eINSIGHT_TYPE_ANOMALY_DETECTION\x10\x04*\x96\x01\n\x10\x43OAApplicability\x12!\n\x1d\x43OA_APPLICABILITY_UNSPECIFIED\x10\x00\x12\x1e\n\x1a\x43OA_APPLICABILITY_PERSONAL\x10\x01\x12\x1e\n\x1a\x43OA_APPLICABILITY_BUSINESS\x10\x02\x12\x1f\n\x1b\x43OA_APPLICABILITY_UNIVERSAL\x10\x03*\x8d\x01\n\x10\x43OAReleaseStatus\x12\"\n\x1e\x43OA_RELEASE_STATUS_UNSPECIFIED\x10\x00\x12\x1b\n\x17\x43OA_RELEASE_STATUS_BETA\x10\x01\x12\x1b\n\x17\x43OA_RELEASE_STATUS_TEST\x10\x02\x12\x1b\n\x17\x43OA_RELEASE_STATUS_LIVE\x10\x03*{\n\x0cTaxonomyType\x12\x1d\n\x19TAXONOMY_TYPE_UNSPECIFIED\x10\x00\x12\"\n\x1eTAXONOMY_TYPE_BANKING_ACCOUNTS\x10\x01\x12(\n$TAXONOMY_TYPE_TRANSACTION_CATEGORIES\x10\x02*\xa5\x02\n\x11GoalOperationType\x12#\n\x1fGOAL_OPERATION_TYPE_UNSPECIFIED\x10\x00\x12\x1e\n\x1aGOAL_OPERATION_TYPE_CREATE\x10\x01\x12\x1e\n\x1aGOAL_OPERATION_TYPE_UPDATE\x10\x02\x12\x1e\n\x1aGOAL_OPERATION_TYPE_DELETE\x10\x03\x12 \n\x1cGOAL_OPERATION_TYPE_COMPLETE\x10\x04\x12\x1d\n\x19GOAL_OPERATION_TYPE_PAUSE\x10\x05\x12\x1e\n\x1aGOAL_OPERATION_TYPE_RESUME\x10\x06\x12*\n&GOAL_OPERATION_TYPE_ALLOCATION_TRIGGER\x10\x07*\x8d\x01\n\nGoalStatus\x12\x1b\n\x17GOAL_STATUS_UNSPECIFIED\x10\x00\x12\x16\n\x12GOAL_STATUS_ACTIVE\x10\x01\x12\x16\n\x12GOAL_STATUS_PAUSED\x10\x02\x12\x19\n\x15GOAL_STATUS_COMPLETED\x10\x03\x12\x17\n\x13GOAL_STATUS_DELETED\x10\x04*\xa9\x01\n\x0fGoalPerformance\x12 \n\x1cGOAL_PERFORMANCE_UNSPECIFIED\x10\x00\x12\x1d\n\x19GOAL_PERFORMANCE_ON_TRACK\x10\x01\x12\x1a\n\x16GOAL_PERFORMANCE_AHEAD\x10\x02\x12\x1b\n\x17GOAL_PERFORMANCE_BEHIND\x10\x03\x12\x1c\n\x18GOAL_PERFORMANCE_AT_RISK\x10\x04*s\n\x0e\x41llocationType\x12\x1f\n\x1b\x41LLOCATION_TYPE_UNSPECIFIED\x10\x00\x12 \n\x1c\x41LLOCATION_TYPE_FIXED_AMOUNT\x10\x01\x12\x1e\n\x1a\x41LLOCATION_TYPE_PERCENTAGE\x10\x02*\xe6\x01\n\x13\x41llocationFrequency\x12$\n ALLOCATION_FREQUENCY_UNSPECIFIED\x10\x00\x12\x1f\n\x1b\x41LLOCATION_FREQUENCY_WEEKLY\x10\x01\x12!\n\x1d\x41LLOCATION_FREQUENCY_BIWEEKLY\x10\x02\x12 \n\x1c\x41LLOCATION_FREQUENCY_MONTHLY\x10\x03\x12\"\n\x1e\x41LLOCATION_FREQUENCY_QUARTERLY\x10\x04\x12\x1f\n\x1b\x41LLOCATION_FREQUENCY_CUSTOM\x10\x05*\xc9\x01\n\rGoalEntryType\x12\x1f\n\x1bGOAL_ENTRY_TYPE_UNSPECIFIED\x10\x00\x12\x1b\n\x17GOAL_ENTRY_TYPE_EARMARK\x10\x01\x12\x1b\n\x17GOAL_ENTRY_TYPE_EXPENSE\x10\x02\x12\x1a\n\x16GOAL_ENTRY_TYPE_REFUND\x10\x03\x12$\n GOAL_ENTRY_TYPE_OPEN_BALANCE_ADJ\x10\x04\x12\x1b\n\x17GOAL_ENTRY_TYPE_REVERSE\x10\x05*\xb4\x01\n\x10SuggestionStatus\x12!\n\x1dSUGGESTION_STATUS_UNSPECIFIED\x10\x00\x12\x1d\n\x19SUGGESTION_STATUS_PENDING\x10\x01\x12\x1e\n\x1aSUGGESTION_STATUS_ACCEPTED\x10\x02\x12\x1f\n\x1bSUGGESTION_STATUS_DISMISSED\x10\x03\x12\x1d\n\x19SUGGESTION_STATUS_EXPIRED\x10\x04*\x8e\x01\n\nDataStatus\x12\x1b\n\x17\x44\x41TA_STATUS_UNSPECIFIED\x10\x00\x12\x19\n\x15\x44\x41TA_STATUS_AVAILABLE\x10\x01\x12\x17\n\x13\x44\x41TA_STATUS_POSTING\x10\x02\x12\x16\n\x12\x44\x41TA_STATUS_POSTED\x10\x03\x12\x17\n\x13\x44\x41TA_STATUS_PARTIAL\x10\x04*\xb8\x01\n\x0cPostingState\x12\x1d\n\x19POSTING_STATE_UNSPECIFIED\x10\x00\x12\x19\n\x15POSTING_STATE_PENDING\x10\x01\x12\x1d\n\x19POSTING_STATE_IN_PROGRESS\x10\x02\x12\x1a\n\x16POSTING_STATE_COMPLETE\x10\x03\x12\x18\n\x14POSTING_STATE_FAILED\x10\x04\x12\x19\n\x15POSTING_STATE_PARTIAL\x10\x05*\xdb\x01\n\x11\x46XRateRequestType\x12$\n FX_RATE_REQUEST_TYPE_UNSPECIFIED\x10\x00\x12&\n\"FX_RATE_REQUEST_TYPE_FETCH_CURRENT\x10\x01\x12$\n FX_RATE_REQUEST_TYPE_FETCH_AS_OF\x10\x02\x12)\n%FX_RATE_REQUEST_TYPE_FETCH_HISTORICAL\x10\x03\x12\'\n#FX_RATE_REQUEST_TYPE_ADD_CURRENCIES\x10\x04*\xd3\x01\n\x14\x46XRateResponseStatus\x12\'\n#FX_RATE_RESPONSE_STATUS_UNSPECIFIED\x10\x00\x12#\n\x1f\x46X_RATE_RESPONSE_STATUS_SUCCESS\x10\x01\x12#\n\x1f\x46X_RATE_RESPONSE_STATUS_PARTIAL\x10\x02\x12!\n\x1d\x46X_RATE_RESPONSE_STATUS_ERROR\x10\x03\x12%\n!FX_RATE_RESPONSE_STATUS_NOT_FOUND\x10\x04*\xc5\r\n\x0bMessageType\x12\x1c\n\x18MESSAGE_TYPE_UNSPECIFIED\x10\x00\x12,\n(MESSAGE_TYPE_JOB_TRANSACTION_DATA_STAGED\x10\x01\x12)\n%MESSAGE_TYPE_JOB_ACCOUNTS_DATA_STAGED\x10\x02\x12,\n(MESSAGE_TYPE_JOB_LIABILITIES_DATA_STAGED\x10\x03\x12,\n(MESSAGE_TYPE_JOB_INVESTMENTS_DATA_STAGED\x10\x04\x12\x1f\n\x1bMESSAGE_TYPE_ETL_COMPLETION\x10\x07\x12\x1b\n\x17MESSAGE_TYPE_ETL_FAILED\x10\x08\x12\"\n\x1eMESSAGE_TYPE_POSTING_COMPLETED\x10\t\x12\x1f\n\x1bMESSAGE_TYPE_POSTING_FAILED\x10\n\x12\"\n\x1eMESSAGE_TYPE_PUSH_NOTIFICATION\x10\x0c\x12\x1b\n\x17MESSAGE_TYPE_COA_UPDATE\x10\x12\x12 \n\x1cMESSAGE_TYPE_WEBHOOK_INGRESS\x10\x19\x12\x1e\n\x1aMESSAGE_TYPE_GENERIC_EVENT\x10\x1a\x12\x1b\n\x17MESSAGE_TYPE_GOAL_EVENT\x10\x1b\x12\x1b\n\x17MESSAGE_TYPE_DATA_READY\x10\x1e\x12%\n!MESSAGE_TYPE_GOAL_EARMARK_REQUEST\x10\x1f\x12#\n\x1fMESSAGE_TYPE_GOAL_SPEND_REQUEST\x10 \x12&\n\"MESSAGE_TYPE_GOAL_ALLOCATE_REQUEST\x10!\x12$\n MESSAGE_TYPE_GOAL_REFUND_REQUEST\x10\"\x12&\n\"MESSAGE_TYPE_GOAL_EARMARK_RESPONSE\x10#\x12$\n MESSAGE_TYPE_GOAL_SPEND_RESPONSE\x10$\x12\'\n#MESSAGE_TYPE_GOAL_ALLOCATE_RESPONSE\x10%\x12%\n!MESSAGE_TYPE_GOAL_REFUND_RESPONSE\x10&\x12\'\n#MESSAGE_TYPE_GOAL_SUBLEDGER_WRITTEN\x10\'\x12(\n$MESSAGE_TYPE_GOAL_SUGGESTION_CREATED\x10(\x12)\n%MESSAGE_TYPE_GOAL_SUGGESTION_RESOLVED\x10)\x12%\n!MESSAGE_TYPE_INSIGHT_OLAP_TRIGGER\x10*\x12!\n\x1dMESSAGE_TYPE_BACKFILL_TRIGGER\x10+\x12\x1f\n\x1bMESSAGE_TYPE_DATA_AVAILABLE\x10,\x12(\n$MESSAGE_TYPE_JOB_RECURRING_TX_STAGED\x10/\x12\x1f\n\x1bMESSAGE_TYPE_FXRATE_REQUEST\x10\x30\x12 \n\x1cMESSAGE_TYPE_FXRATE_RESPONSE\x10\x31\x12/\n+MESSAGE_TYPE_CATEGORY_DETERMINATION_REQUEST\x10\x32\x12\x30\n,MESSAGE_TYPE_CATEGORY_DETERMINATION_RESPONSE\x10\x33\x12\x33\n/MESSAGE_TYPE_POSTING_RULE_DETERMINATION_REQUEST\x10\x34\x12\x34\n0MESSAGE_TYPE_POSTING_RULE_DETERMINATION_RESPONSE\x10\x35\x12)\n%MESSAGE_TYPE_INSIGHT_ANALYSIS_REQUEST\x10\x36\x12*\n&MESSAGE_TYPE_INSIGHT_ANALYSIS_RESPONSE\x10\x37\x12/\n+MESSAGE_TYPE_WEBHOOK_CONFIG_UPDATE_DETECTED\x10\x38\x12!\n\x1dMESSAGE_TYPE_TAXONOMY_RELEASE\x10\x39\x12&\n\"MESSAGE_TYPE_POSTING_RULES_RELEASE\x10:\x12&\n\"MESSAGE_TYPE_CONNECTION_RECONCILED\x10;\"\x04\x08\x05\x10\x05\"\x04\x08\x06\x10\x06\"\x04\x08\x0b\x10\x0b\"\x04\x08\r\x10\r\"\x04\x08\x13\x10\x13\"\x04\x08\x14\x10\x14\"\x04\x08\x15\x10\x15\"\x04\x08\x16\x10\x16\"\x04\x08\x18\x10\x18\"\x04\x08\x1c\x10\x1c\"\x04\x08\x1d\x10\x1d\"\x04\x08-\x10-\"\x04\x08.\x10.BDZBgithub.com/LetA-Tech/mcfo-neonlink/gen/go/messaging/v1;messagingv1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'messaging.v1.messaging_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'ZBgithub.com/LetA-Tech/mcfo-neonlink/gen/go/messaging/v1;messagingv1'
  _globals['_BACKFILLTRIGGERCONTEXT_METADATAENTRY']._loaded_options = None
  _globals['_BACKFILLTRIGGERCONTEXT_METADATAENTRY']._serialized_options = b'8\001'
  _globals['_WEBHOOKINGRESSCONTEXT_HEADERSENTRY']._loaded_options = None
  _globals['_WEBHOOKINGRESSCONTEXT_HEADERSENTRY']._serialized_options = b'8\001'
  _globals['_GOALOPERATIONRESPONSE_AGGREGATESBYGOALENTRY']._loaded_options = None
  _globals['_GOALOPERATIONRESPONSE_AGGREGATESBYGOALENTRY']._serialized_options = b'8\001'
  _globals['_INSIGHTTRIGGERCONTEXT_HINTSENTRY']._loaded_options = None
  _globals['_INSIGHTTRIGGERCONTEXT_HINTSENTRY']._serialized_options = b'8\001'
  _globals['_GENERICEVENTCONTEXT_ATTRIBUTESENTRY']._loaded_options = None
  _globals['_GENERICEVENTCONTEXT_ATTRIBUTESENTRY']._serialized_options = b'8\001'
  _globals['_DATAQUALITYMETRICS_FIELDCOMPLETENESSENTRY']._loaded_options = None
  _globals['_DATAQUALITYMETRICS_FIELDCOMPLETENESSENTRY']._serialized_options = b'8\001'
  _globals['_DATAQUALITYMETRICS_VALIDATIONRESULTSENTRY']._loaded_options = None
  _globals['_DATAQUALITYMETRICS_VALIDATIONRESULTSENTRY']._serialized_options = b'8\001'
  _globals['_EARMARKREQUEST_SOURCEMETADATAENTRY']._loaded_options = None
  _globals['_EARMARKREQUEST_SOURCEMETADATAENTRY']._serialized_options = b'8\001'
  _globals['_PROCESSINGSTATUS']._serialized_start=13734
  _globals['_PROCESSINGSTATUS']._serialized_end=14012
  _globals['_DATATYPE']._serialized_start=14015
  _globals['_DATATYPE']._serialized_end=14415
  _globals['_SOURCESERVICE']._serialized_start=14418
  _globals['_SOURCESERVICE']._serialized_end=15208
  _globals['_TARGETSERVICE']._serialized_start=15211
  _globals['_TARGETSERVICE']._serialized_end=15640
  _globals['_SYNCTYPE']._serialized_start=15643
  _globals['_SYNCTYPE']._serialized_end=15819
  _globals['_ENTRYSIDE']._serialized_start=15821
  _globals['_ENTRYSIDE']._serialized_end=15905
  _globals['_DETERMINATIONTYPE']._serialized_start=15908
  _globals['_DETERMINATIONTYPE']._serialized_end=16071
  _globals['_INSIGHTTYPE']._serialized_start=16074
  _globals['_INSIGHTTYPE']._serialized_end=16259
  _globals['_COAAPPLICABILITY']._serialized_start=16262
  _globals['_COAAPPLICABILITY']._serialized_end=16412
  _globals['_COARELEASESTATUS']._serialized_start=16415
  _globals['_COARELEASESTATUS']._serialized_end=16556
  _globals['_TAXONOMYTYPE']._serialized_start=16558
  _globals['_TAXONOMYTYPE']._serialized_end=16681
  _globals['_GOALOPERATIONTYPE']._serialized_start=16684
  _globals['_GOALOPERATIONTYPE']._serialized_end=16977
  _globals['_GOALSTATUS']._serialized_start=16980
  _globals['_GOALSTATUS']._serialized_end=17121
  _globals['_GOALPERFORMANCE']._serialized_start=17124
  _globals['_GOALPERFORMANCE']._serialized_end=17293
  _globals['_ALLOCATIONTYPE']._serialized_start=17295
  _globals['_ALLOCATIONTYPE']._serialized_end=17410
  _globals['_ALLOCATIONFREQUENCY']._serialized_start=17413
  _globals['_ALLOCATIONFREQUENCY']._serialized_end=17643
  _globals['_GOALENTRYTYPE']._serialized_start=17646
  _globals['_GOALENTRYTYPE']._serialized_end=17847
  _globals['_SUGGESTIONSTATUS']._serialized_start=17850
  _globals['_SUGGESTIONSTATUS']._serialized_end=18030
  _globals['_DATASTATUS']._serialized_start=18033
  _globals['_DATASTATUS']._serialized_end=18175
  _globals['_POSTINGSTATE']._serialized_start=18178
  _globals['_POSTINGSTATE']._serialized_end=18362
  _globals['_FXRATEREQUESTTYPE']._serialized_start=18365
  _globals['_FXRATEREQUESTTYPE']._serialized_end=18584
  _globals['_FXRATERESPONSESTATUS']._serialized_start=18587
  _globals['_FXRATERESPONSESTATUS']._serialized_end=18798
  _globals['_MESSAGETYPE']._serialized_start=18801
  _globals['_MESSAGETYPE']._serialized_end=20534
  _globals['_JOBDISPATCH']._serialized_start=89
  _globals['_JOBDISPATCH']._serialized_end=2542
  _globals['_ETLJOBCONTEXT']._serialized_start=2545
  _globals['_ETLJOBCONTEXT']._serialized_end=2771
  _globals['_ETLCOMPLETIONCONTEXT']._serialized_start=2774
  _globals['_ETLCOMPLETIONCONTEXT']._serialized_end=2954
  _globals['_POSTINGCOMPLETIONCONTEXT']._serialized_start=2957
  _globals['_POSTINGCOMPLETIONCONTEXT']._serialized_end=3154
  _globals['_BACKFILLTRIGGERCONTEXT']._serialized_start=3157
  _globals['_BACKFILLTRIGGERCONTEXT']._serialized_end=3633
  _globals['_BACKFILLTRIGGERCONTEXT_METADATAENTRY']._serialized_start=3558
  _globals['_BACKFILLTRIGGERCONTEXT_METADATAENTRY']._serialized_end=3605
  _globals['_WEBHOOKINGRESSCONTEXT']._serialized_start=3636
  _globals['_WEBHOOKINGRESSCONTEXT']._serialized_end=3939
  _globals['_WEBHOOKINGRESSCONTEXT_HEADERSENTRY']._serialized_start=3893
  _globals['_WEBHOOKINGRESSCONTEXT_HEADERSENTRY']._serialized_end=3939
  _globals['_GOALOPERATIONREQUEST']._serialized_start=3942
  _globals['_GOALOPERATIONREQUEST']._serialized_end=4267
  _globals['_GOALOPERATIONRESPONSE']._serialized_start=4270
  _globals['_GOALOPERATIONRESPONSE']._serialized_end=4645
  _globals['_GOALOPERATIONRESPONSE_AGGREGATESBYGOALENTRY']._serialized_start=4551
  _globals['_GOALOPERATIONRESPONSE_AGGREGATESBYGOALENTRY']._serialized_end=4645
  _globals['_GOALSUBLEDGEREVENT']._serialized_start=4648
  _globals['_GOALSUBLEDGEREVENT']._serialized_end=4859
  _globals['_INSIGHTTRIGGERCONTEXT']._serialized_start=4862
  _globals['_INSIGHTTRIGGERCONTEXT']._serialized_end=5165
  _globals['_INSIGHTTRIGGERCONTEXT_HINTSENTRY']._serialized_start=5121
  _globals['_INSIGHTTRIGGERCONTEXT_HINTSENTRY']._serialized_end=5165
  _globals['_DATAREADYCONTEXT']._serialized_start=5168
  _globals['_DATAREADYCONTEXT']._serialized_end=5394
  _globals['_CATEGORYDETERMINATIONREQUESTCONTEXT']._serialized_start=5396
  _globals['_CATEGORYDETERMINATIONREQUESTCONTEXT']._serialized_end=5472
  _globals['_CATEGORYDETERMINATIONRESPONSECONTEXT']._serialized_start=5475
  _globals['_CATEGORYDETERMINATIONRESPONSECONTEXT']._serialized_end=5664
  _globals['_POSTINGRULEDETERMINATIONREQUESTCONTEXT']._serialized_start=5667
  _globals['_POSTINGRULEDETERMINATIONREQUESTCONTEXT']._serialized_end=5816
  _globals['_POSTINGRULEDETERMINATIONRESPONSECONTEXT']._serialized_start=5819
  _globals['_POSTINGRULEDETERMINATIONRESPONSECONTEXT']._serialized_end=6011
  _globals['_INSIGHTANALYSISREQUESTCONTEXT']._serialized_start=6014
  _globals['_INSIGHTANALYSISREQUESTCONTEXT']._serialized_end=6221
  _globals['_INSIGHTANALYSISRESPONSECONTEXT']._serialized_start=6224
  _globals['_INSIGHTANALYSISRESPONSECONTEXT']._serialized_end=6419
  _globals['_TAXONOMYRELEASECONTEXT']._serialized_start=6422
  _globals['_TAXONOMYRELEASECONTEXT']._serialized_end=6681
  _globals['_POSTINGRULESRELEASECONTEXT']._serialized_start=6684
  _globals['_POSTINGRULESRELEASECONTEXT']._serialized_end=6907
  _globals['_COAUPDATECONTEXT']._serialized_start=6910
  _globals['_COAUPDATECONTEXT']._serialized_end=7123
  _globals['_GENERICEVENTCONTEXT']._serialized_start=7126
  _globals['_GENERICEVENTCONTEXT']._serialized_end=7427
  _globals['_GENERICEVENTCONTEXT_ATTRIBUTESENTRY']._serialized_start=7378
  _globals['_GENERICEVENTCONTEXT_ATTRIBUTESENTRY']._serialized_end=7427
  _globals['_GOALEVENTCONTEXT']._serialized_start=7430
  _globals['_GOALEVENTCONTEXT']._serialized_end=7651
  _globals['_SUGGESTIONEVENT']._serialized_start=7654
  _globals['_SUGGESTIONEVENT']._serialized_end=7974
  _globals['_DATAAVAILABLECONTEXT']._serialized_start=7977
  _globals['_DATAAVAILABLECONTEXT']._serialized_end=8241
  _globals['_ETLFAILEDCONTEXT']._serialized_start=8244
  _globals['_ETLFAILEDCONTEXT']._serialized_end=8445
  _globals['_POSTINGFAILEDCONTEXT']._serialized_start=8448
  _globals['_POSTINGFAILEDCONTEXT']._serialized_end=8632
  _globals['_FXRATEREQUESTCONTEXT']._serialized_start=8635
  _globals['_FXRATEREQUESTCONTEXT']._serialized_end=8939
  _globals['_FXRATERESPONSECONTEXT']._serialized_start=8942
  _globals['_FXRATERESPONSECONTEXT']._serialized_end=9151
  _globals['_CONNECTIONRECONCILEDCONTEXT']._serialized_start=9154
  _globals['_CONNECTIONRECONCILEDCONTEXT']._serialized_end=9408
  _globals['_DUPLICATEACCOUNTPAIR']._serialized_start=9410
  _globals['_DUPLICATEACCOUNTPAIR']._serialized_end=9486
  _globals['_IDENTITYCONTEXT']._serialized_start=9489
  _globals['_IDENTITYCONTEXT']._serialized_end=9779
  _globals['_MESSAGEHEADER']._serialized_start=9782
  _globals['_MESSAGEHEADER']._serialized_end=10129
  _globals['_PROCESSINGSUMMARY']._serialized_start=10132
  _globals['_PROCESSINGSUMMARY']._serialized_end=10358
  _globals['_DATAQUALITYMETRICS']._serialized_start=10361
  _globals['_DATAQUALITYMETRICS']._serialized_end=10790
  _globals['_DATAQUALITYMETRICS_FIELDCOMPLETENESSENTRY']._serialized_start=10676
  _globals['_DATAQUALITYMETRICS_FIELDCOMPLETENESSENTRY']._serialized_end=10732
  _globals['_DATAQUALITYMETRICS_VALIDATIONRESULTSENTRY']._serialized_start=10734
  _globals['_DATAQUALITYMETRICS_VALIDATIONRESULTSENTRY']._serialized_end=10790
  _globals['_POSTINGSUMMARY']._serialized_start=10793
  _globals['_POSTINGSUMMARY']._serialized_end=11075
  _globals['_DATADETAILS']._serialized_start=11077
  _globals['_DATADETAILS']._serialized_end=11128
  _globals['_POSTINGLEG']._serialized_start=11131
  _globals['_POSTINGLEG']._serialized_end=11278
  _globals['_GOALEVENTPAYLOAD']._serialized_start=11281
  _globals['_GOALEVENTPAYLOAD']._serialized_end=11721
  _globals['_ALLOCATIONSETTINGS']._serialized_start=11724
  _globals['_ALLOCATIONSETTINGS']._serialized_end=12027
  _globals['_ALLOCATIONCONTEXT']._serialized_start=12030
  _globals['_ALLOCATIONCONTEXT']._serialized_end=12287
  _globals['_EARMARKREQUEST']._serialized_start=12290
  _globals['_EARMARKREQUEST']._serialized_end=12501
  _globals['_EARMARKREQUEST_SOURCEMETADATAENTRY']._serialized_start=12448
  _globals['_EARMARKREQUEST_SOURCEMETADATAENTRY']._serialized_end=12501
  _globals['_SPENDREQUEST']._serialized_start=12503
  _globals['_SPENDREQUEST']._serialized_end=12611
  _globals['_ALLOCATEREQUEST']._serialized_start=12614
  _globals['_ALLOCATEREQUEST']._serialized_end=12751
  _globals['_ALLOCATION']._serialized_start=12753
  _globals['_ALLOCATION']._serialized_end=12812
  _globals['_REFUNDREQUEST']._serialized_start=12815
  _globals['_REFUNDREQUEST']._serialized_end=12951
  _globals['_SUBLEDGERENTRY']._serialized_start=12954
  _globals['_SUBLEDGERENTRY']._serialized_end=13239
  _globals['_GOALAGGREGATES']._serialized_start=13242
  _globals['_GOALAGGREGATES']._serialized_end=13463
  _globals['_SUGGESTIONCANDIDATE']._serialized_start=13465
  _globals['_SUGGESTIONCANDIDATE']._serialized_end=13576
  _globals['_FXRATEENTRY']._serialized_start=13579
  _globals['_FXRATEENTRY']._serialized_end=13731
# @@protoc_insertion_point(module_scope)



"""Kafka consumer with manual offset commit, failure tracking, and panic recovery."""

from __future__ import annotations

import logging
import math
import random
import threading
import time
from datetime import datetime, timezone
from typing import Callable, Optional, Protocol

from confluent_kafka import (
    Consumer as ConfluentConsumer,
    KafkaError,
    KafkaException,
    Message,
    TopicPartition,
    TIMESTAMP_CREATE_TIME,
    TIMESTAMP_LOG_APPEND_TIME,
)

from neonlink.circuit_breaker import CircuitBreaker
from neonlink.config import Config
from neonlink.errors import PermanentError, is_permanent
from neonlink.metrics import NoopMetrics
from neonlink.record import HEADER_RETRY_COUNT, Record
from neonlink.tracing import end_span_with_error, extract_trace_context, start_consumer_span

try:
    from opentelemetry import context as otel_context
    _HAS_OTEL = True
except ImportError:
    _HAS_OTEL = False

logger = logging.getLogger(__name__)


class MessageHandler(Protocol):
    """Protocol for message handlers."""

    def handle_message(self, record: Record) -> None:
        """Process a single record.

        Return normally to acknowledge (commit offset).
        Raise an exception to skip commit — Kafka redelivers on next poll.
        """
        ...


class MessageHandlerWithCommit(Protocol):
    """Protocol for handlers that manually control Kafka offset commits."""

    def handle_message_with_commit(
        self,
        record: Record,
        commit: Callable[[], bool],
    ) -> None:
        """Process a record and call commit() only after durable side-effects."""
        ...


class _FailureTracker:
    """In-memory per-record failure counter.

    Fixes the broken retry_count header problem: Kafka redelivers the same
    record with the same headers, so the retry_count header never advances
    unless the producer re-publishes. This tracker maintains a local count
    keyed by topic:partition:offset.
    """

    def __init__(self) -> None:
        self._counts: dict[str, int] = {}
        self._lock = threading.Lock()

    @staticmethod
    def _key(topic: str, partition: int, offset: int) -> str:
        return f"{topic}:{partition}:{offset}"

    def increment(self, topic: str, partition: int, offset: int) -> int:
        with self._lock:
            k = self._key(topic, partition, offset)
            self._counts[k] = self._counts.get(k, 0) + 1
            return self._counts[k]

    def remove(self, topic: str, partition: int, offset: int) -> None:
        with self._lock:
            self._counts.pop(self._key(topic, partition, offset), None)

    def clear_partition(self, topic: str, partition: int) -> None:
        with self._lock:
            prefix = f"{topic}:{partition}:"
            to_delete = [k for k in self._counts if k.startswith(prefix)]
            for k in to_delete:
                del self._counts[k]


def _consumer_poll_backoff(consecutive_errors: int, initial_wait_ms: int, max_wait_ms: int) -> float:
    """Calculate backoff in seconds for consumer poll retries."""
    if consecutive_errors <= 0:
        return 0.0
    base = initial_wait_ms * math.pow(2, consecutive_errors - 1)
    if base > max_wait_ms:
        base = max_wait_ms
    # Jitter: 75-125% of base.
    jittered = base * (0.75 + random.random() * 0.5)
    return jittered / 1000.0  # Convert to seconds


class Consumer:
    """Subscribes to Kafka topics and dispatches records to a MessageHandler.

    On handler success: offset is committed.
    On handler error: offset is NOT committed. Kafka redelivers on next poll.
    If retry_count >= poison pill threshold: offset is committed (skip past).

    Features:
    - In-memory failure tracking (fixes broken retry_count header)
    - Exponential backoff on poll errors (prevents hot-spin)
    - Handler panic recovery (consumer continues after handler exceptions)
    - Rebalance-aware failure tracker cleanup
    """

    def __init__(self, cfg: Config, topics: list[str]) -> None:
        cfg.validate()
        if not cfg.consumer_group:
            raise ValueError("neonlink: consumer group is required")
        if not topics:
            raise ValueError("neonlink: at least one topic is required")

        self._consumer = ConfluentConsumer(cfg.to_consumer_config())
        self._topics = topics
        self._cfg = cfg
        self._tracker = _FailureTracker()
        self._metrics = cfg.metrics if cfg.metrics is not None else NoopMetrics()
        self._cb = CircuitBreaker(
            name="consumer",
            max_failures=cfg.cb_max_failures,
            reset_timeout_sec=cfg.cb_reset_timeout_sec,
            on_state_change=lambda n, f, t: self._metrics.circuit_breaker_state_change(n, f, t),
        )
        self._stop_event = threading.Event()
        self._cancel_event: Optional[threading.Event] = None
        self._started = False
        self._closed = False
        self._last_poll_error = False
        self._dlq_producer = None
        self._revoked_partitions: set[str] = set()
        self._revoked_lock = threading.Lock()

    def subscribe(
        self,
        handler: MessageHandler,
        *,
        cancel_event: Optional[threading.Event] = None,
    ) -> None:
        """Start the consume loop, blocking until stop() is called.

        Args:
            handler: MessageHandler implementation to process each record.
            cancel_event: Optional threading.Event that, when set, stops the loop.
        """
        if self._started:
            raise RuntimeError("neonlink: consumer already started")
        self._started = True

        if cancel_event is not None:
            self._cancel_event = cancel_event

        def _on_revoke(consumer, partitions):
            with self._revoked_lock:
                for tp in partitions:
                    self._revoked_partitions.add(f"{tp.topic}:{tp.partition}")
                    self._tracker.clear_partition(tp.topic, tp.partition)
                    logger.info(
                        "neonlink: partition revoked topic=%s partition=%d",
                        tp.topic, tp.partition,
                    )
            # NOTE: Do NOT call consumer.commit() here. Per-message commits
            # (self._commit(msg)) are synchronous, so acknowledged offsets are
            # already persisted. A blanket commit() would advance the position
            # past records whose handlers failed, defeating the "don't commit
            # on error" guarantee. The Go SDK uses CommitMarkedOffsets which
            # only flushes explicitly marked offsets — this is the equivalent.

        def _on_assign(consumer, partitions):
            with self._revoked_lock:
                for tp in partitions:
                    self._revoked_partitions.discard(f"{tp.topic}:{tp.partition}")
                    logger.info(
                        "neonlink: partition assigned topic=%s partition=%d",
                        tp.topic, tp.partition,
                    )

        self._consumer.subscribe(
            self._topics,
            on_revoke=_on_revoke,
            on_assign=_on_assign,
        )
        logger.info(
            "neonlink: consumer starting, group=%s topics=%s",
            self._cfg.consumer_group,
            self._topics,
        )

        consecutive_errors = 0

        try:
            while not self._is_stopped():
                msg = self._poll_one()
                if msg is None:
                    # _poll_one returns None on error or timeout.
                    # Check if it was an error (CB or poll error).
                    if self._last_poll_error:
                        consecutive_errors += 1
                        backoff = _consumer_poll_backoff(
                            consecutive_errors,
                            self._cfg.retry_initial_wait_ms,
                            self._cfg.retry_max_wait_ms,
                        )
                        logger.warning(
                            "neonlink: poll error, backing off "
                            "consecutive_errors=%d backoff_sec=%.2f",
                            consecutive_errors, backoff,
                        )
                        if self._wait_for_stop(timeout=backoff):
                            break  # Stop requested during backoff
                        self._last_poll_error = False
                    continue

                consecutive_errors = 0
                record = _from_confluent_message(msg)

                # Rebalance fencing: skip records from revoked partitions.
                with self._revoked_lock:
                    if f"{record.topic}:{record.partition}" in self._revoked_partitions:
                        logger.debug(
                            "neonlink: skipping record from revoked partition "
                            "topic=%s partition=%s offset=%s",
                            record.topic, record.partition, record.offset,
                        )
                        continue

                # Consumer lag metrics.
                self._emit_consumer_lag(record)

                # Extract OTel trace context for distributed tracing continuity.
                otel_token = _attach_trace_context(record)

                # Start consumer span for observability.
                span = start_consumer_span(record)

                # Panic recovery: catch handler exceptions.
                handler_start = time.monotonic()
                handler_err = None
                commit_handler = getattr(handler, "handle_message_with_commit", None)
                has_commit_control = callable(commit_handler)
                try:
                    if has_commit_control:
                        commit_handler(record, lambda: self._commit(msg))
                    else:
                        handler.handle_message(record)
                except Exception as exc:
                    handler_err = exc
                    self._metrics.consume_error(record.topic, record.partition, exc)
                    self._handle_processing_error(record, msg, exc)
                finally:
                    end_span_with_error(span, handler_err)
                    _detach_trace_context(otel_token)

                if handler_err is not None:
                    continue

                # Success: commit and clear failure tracker.
                self._metrics.consume_success(
                    record.topic, record.partition, time.monotonic() - handler_start
                )
                self._tracker.remove(record.topic, record.partition, record.offset)
                if not has_commit_control:
                    self._commit(msg)
        finally:
            self._consumer.close()
            logger.info("neonlink: consumer stopped")

    def subscribe_with_dlq(
        self,
        handler: MessageHandler,
        dlq_producer,
        *,
        cancel_event: Optional[threading.Event] = None,
    ) -> None:
        """Start the consume loop with DLQ routing for poison pills.

        When a poison pill is detected, the record is routed to the DLQ before
        committing the offset. If DLQ routing fails, the offset is NOT committed.
        """
        self._dlq_producer = dlq_producer
        self.subscribe(handler, cancel_event=cancel_event)

    def stop(self) -> None:
        """Request graceful shutdown of the consume loop."""
        self._stop_event.set()
        if self._cancel_event is not None:
            self._cancel_event.set()

    def close(self) -> None:
        """Stop and close the consumer."""
        if self._closed:
            return
        self._closed = True
        self.stop()

    def pause_fetch(self, topics: list[str]) -> None:
        """Pause fetching from the specified topics for custom backpressure."""
        partitions = self._consumer.assignment()
        to_pause = [tp for tp in partitions if tp.topic in topics]
        if to_pause:
            self._consumer.pause(to_pause)
            logger.info("neonlink: fetch paused topics=%s", topics)

    def resume_fetch(self, topics: list[str]) -> None:
        """Resume fetching from the specified topics."""
        partitions = self._consumer.assignment()
        to_resume = [tp for tp in partitions if tp.topic in topics]
        if to_resume:
            self._consumer.resume(to_resume)
            logger.info("neonlink: fetch resumed topics=%s", topics)

    def _emit_consumer_lag(self, record: Record) -> None:
        """Emit consumer lag metric for the record's partition (best-effort)."""
        try:
            tp = TopicPartition(record.topic, record.partition)
            _, high = self._consumer.get_watermark_offsets(tp, cached=True)
            lag = high - record.offset - 1
            if lag < 0:
                lag = 0
            self._metrics.consumer_lag(record.topic, record.partition, lag)
        except Exception:
            pass

    def _poll_one(self) -> Optional[Message]:
        """Poll for a single message with circuit breaker."""
        msg: Optional[Message] = None
        self._last_poll_error = False

        def _do_poll() -> None:
            nonlocal msg
            msg = self._consumer.poll(timeout=1.0)
            if msg is None:
                return
            err = msg.error()
            if err is None:
                return
            if err.code() == KafkaError._PARTITION_EOF:
                msg = None
                return
            raise KafkaException(err)

        try:
            self._cb.execute_with_classifier(
                _do_poll,
                should_trip=lambda err: not is_permanent(err),
            )
        except Exception as exc:
            logger.warning("neonlink: poll error: %s", exc)
            self._last_poll_error = True
            return None

        if msg is None:
            return None

        return msg

    def _commit(self, msg: Message) -> bool:
        """Commit offset for the consumed message with bounded retry.

        Returns True on success, False if all retries exhausted.
        """
        topic = msg.topic() or ""
        partition = msg.partition() or 0
        if self._is_partition_revoked(topic, partition):
            logger.warning(
                "neonlink: commit fenced for revoked partition "
                "topic=%s partition=%s offset=%s",
                topic,
                partition,
                msg.offset(),
            )
            return False

        max_retries = 3
        for attempt in range(max_retries):
            if self._is_partition_revoked(topic, partition):
                logger.warning(
                    "neonlink: commit fenced during retry for revoked partition "
                    "topic=%s partition=%s offset=%s",
                    topic,
                    partition,
                    msg.offset(),
                )
                return False
            try:
                self._consumer.commit(message=msg, asynchronous=False)
                return True
            except KafkaException as exc:
                logger.warning(
                    "neonlink: offset commit attempt %d/%d failed "
                    "topic=%s partition=%s offset=%s: %s",
                    attempt + 1,
                    max_retries,
                    msg.topic(),
                    msg.partition(),
                    msg.offset(),
                    exc,
                )
        logger.error(
            "neonlink: offset commit failed after %d retries "
            "topic=%s partition=%s offset=%s",
            max_retries,
            msg.topic(),
            msg.partition(),
            msg.offset(),
        )
        return False

    def _handle_processing_error(
        self, record: Record, msg: Message, exc: Exception
    ) -> None:
        """Decide whether to skip (commit) or leave for redelivery."""
        header_retry_count = _get_retry_count(record)
        tracker_count = self._tracker.increment(record.topic, record.partition, record.offset)

        effective_retry_count = max(header_retry_count, tracker_count)

        if effective_retry_count >= self._cfg.poison_pill_threshold:
            self._metrics.poison_pill_detected(record.topic, record.partition, record.offset)
            logger.error(
                "neonlink: poison pill detected "
                "topic=%s partition=%s offset=%s effective_retry=%d "
                "header_retry=%d tracker=%d: %s",
                record.topic,
                record.partition,
                record.offset,
                effective_retry_count,
                header_retry_count,
                tracker_count,
                exc,
            )

            if self._dlq_producer is not None:
                from neonlink.dlq import route_to_dlq

                try:
                    route_to_dlq(
                        self._dlq_producer,
                        self._cfg.dlq_topic,
                        record,
                        exc,
                        poison_reason="MAX_RETRIES_EXCEEDED",
                        retry_count=effective_retry_count,
                        consumer_group=self._cfg.consumer_group,
                    )
                    self._metrics.dlq_routed(record.topic, exc)
                    logger.warning(
                        "neonlink: poison pill routed to DLQ, committing "
                        "topic=%s offset=%s",
                        record.topic,
                        record.offset,
                    )
                except Exception as dlq_exc:
                    logger.error(
                        "neonlink: DLQ routing failed, offset NOT committed "
                        "topic=%s offset=%s: %s",
                        record.topic,
                        record.offset,
                        dlq_exc,
                    )
                    # Do NOT commit — leave for redelivery so we don't lose data.
                    return
            else:
                # No DLQ configured — do NOT commit, log CRITICAL.
                self._metrics.poison_pill_blocked(record.topic, record.partition, record.offset)
                logger.error(
                    "neonlink: CRITICAL poison pill with no DLQ, offset NOT committed "
                    "topic=%s partition=%s offset=%s",
                    record.topic,
                    record.partition,
                    record.offset,
                )
                return

            if self._commit(msg):
                self._tracker.remove(record.topic, record.partition, record.offset)
            else:
                logger.error(
                    "neonlink: poison pill commit failed after DLQ route "
                    "topic=%s partition=%s offset=%s",
                    record.topic,
                    record.partition,
                    record.offset,
                )
            return

        logger.warning(
            "neonlink: handler error, will redeliver "
            "topic=%s partition=%s offset=%s effective_retry=%d: %s",
            record.topic,
            record.partition,
            record.offset,
            effective_retry_count,
            exc,
        )

    def _is_stopped(self) -> bool:
        """Check if either the internal stop event or external cancel event is set."""
        if self._stop_event.is_set():
            return True
        if self._cancel_event is not None and self._cancel_event.is_set():
            return True
        return False

    def _wait_for_stop(self, timeout: float) -> bool:
        """Wait up to *timeout* seconds for a stop signal.

        Returns True if a stop was requested (i.e. the loop should break).
        Because threading.Event.wait() can only watch a single event, we poll
        in short intervals so we notice either event promptly.
        """
        deadline = time.monotonic() + timeout
        while True:
            remaining = deadline - time.monotonic()
            if remaining <= 0:
                return self._is_stopped()
            # Poll in 0.1 s slices so we notice cancel_event quickly.
            wait_slice = min(remaining, 0.1)
            if self._stop_event.wait(timeout=wait_slice):
                return True
            if self._cancel_event is not None and self._cancel_event.is_set():
                return True
        return False

    def _is_partition_revoked(self, topic: str, partition: int) -> bool:
        with self._revoked_lock:
            return f"{topic}:{partition}" in self._revoked_partitions


def _attach_trace_context(record: Record):
    """Attach OTel trace context from record headers. Returns token or None."""
    if not _HAS_OTEL:
        return None
    trace_ctx = extract_trace_context(record)
    if trace_ctx is not None:
        return otel_context.attach(trace_ctx)
    return None


def _detach_trace_context(token) -> None:
    """Detach OTel trace context if previously attached."""
    if token is not None and _HAS_OTEL:
        otel_context.detach(token)


def _from_confluent_message(msg: Message) -> Record:
    """Convert a confluent-kafka Message to a Record."""
    headers: dict[str, str | bytes] = {}
    raw_headers = msg.headers()
    if raw_headers:
        for key, val in raw_headers:
            if val is not None:
                headers[key] = val  # store raw bytes, preserving binary metadata

    timestamp = None
    ts_type, ts_value = msg.timestamp()
    if ts_type in (TIMESTAMP_CREATE_TIME, TIMESTAMP_LOG_APPEND_TIME) and ts_value > 0:
        timestamp = datetime.fromtimestamp(ts_value / 1000.0, tz=timezone.utc)

    return Record(
        topic=msg.topic() or "",
        key=msg.key(),
        value=msg.value(),
        headers=headers,
        partition=msg.partition() or 0,
        offset=msg.offset() or 0,
        timestamp=timestamp,
    )


def _get_retry_count(record: Record) -> int:
    """Extract retry_count from record headers."""
    val = record.get_header(HEADER_RETRY_COUNT)
    if not val:
        return 0
    try:
        return int(val)
    except ValueError:
        return 0



"""TopicRouter dispatches consumed records to topic-specific handlers.

Implements the MessageHandler protocol so it plugs directly into
Consumer.subscribe().

Usage::

    from neonlink import TopicRouter, Consumer

    router = TopicRouter()
    router.handle("posting-input", posting_handler)
    router.handle_func("sync-triggers", lambda rec: process_sync(rec))

    consumer = Consumer(cfg, topics=router.topics())
    consumer.subscribe(router)
"""

from __future__ import annotations

from typing import Callable

from neonlink.errors import PermanentError
from neonlink.record import Record


class TopicRouter:
    """Routes records to topic-specific handlers.

    Implements MessageHandler so it can be passed directly to Consumer.subscribe().
    """

    def __init__(self) -> None:
        self._handlers: dict[str, object] = {}

    def handle(self, topic: str, handler: object) -> None:
        """Register a MessageHandler for a topic.

        Raises ValueError at startup on empty topic, None handler, or duplicate
        registration — misconfiguration should fail fast.
        """
        if not topic:
            raise ValueError("neonlink: TopicRouter.handle called with empty topic")
        if handler is None:
            raise ValueError("neonlink: TopicRouter.handle called with None handler")
        if not callable(getattr(handler, "handle_message", None)):
            raise TypeError(
                f"neonlink: handler for topic {topic!r} must implement handle_message(record)"
            )
        if topic in self._handlers:
            raise ValueError(
                f"neonlink: TopicRouter.handle called twice for topic {topic!r}"
            )
        self._handlers[topic] = handler

    def handle_func(self, topic: str, fn: Callable[[Record], None]) -> None:
        """Register a plain function as a handler for a topic."""
        if fn is None:
            raise ValueError("neonlink: TopicRouter.handle_func called with None fn")
        self.handle(topic, _FuncHandler(fn))

    def topics(self) -> list[str]:
        """Return registered topic names in sorted order.

        Pass this directly to the Consumer constructor.
        """
        return sorted(self._handlers.keys())

    def handle_message(self, record: Record) -> None:
        """Dispatch the record to the handler registered for record.topic.

        Raises PermanentError if no handler is registered — unroutable records
        should not block the consumer with infinite redelivery.
        """
        handler = self._handlers.get(record.topic)
        if handler is None:
            raise PermanentError(
                f"neonlink: no handler registered for topic {record.topic!r}"
            )
        handler.handle_message(record)


class _FuncHandler:
    """Adapts a plain function to the MessageHandler protocol."""

    __slots__ = ("_fn",)

    def __init__(self, fn: Callable[[Record], None]) -> None:
        self._fn = fn

    def handle_message(self, record: Record) -> None:
        self._fn(record)



"""Tests for neonlink.dlq."""

from datetime import datetime, timezone

from neonlink.dlq import _classify_reason, route_to_dlq
from neonlink.mock import MockProducer
from neonlink.record import HEADER_RETRY_COUNT, Record


class TestClassifyReason:
    def test_with_retry_count(self):
        rec = Record(topic="t", headers={HEADER_RETRY_COUNT: "5"})
        assert _classify_reason(rec) == "MAX_RETRIES_EXCEEDED(retry_count=5)"

    def test_without_retry_count(self):
        rec = Record(topic="t", headers={})
        assert _classify_reason(rec) == "NON_RETRIABLE_ERROR"

    def test_zero_retry_count(self):
        rec = Record(topic="t", headers={HEADER_RETRY_COUNT: "0"})
        assert _classify_reason(rec) == "NON_RETRIABLE_ERROR"

    def test_invalid_retry_count(self):
        rec = Record(topic="t", headers={HEADER_RETRY_COUNT: "bad"})
        assert _classify_reason(rec) == "NON_RETRIABLE_ERROR"


class TestRouteToDLQTimestamp:
    """R-01: DLQ timestamp must have real sub-second precision."""

    def test_timestamp_has_subsecond_precision(self):
        mp = MockProducer()
        rec = Record(topic="src", key=b"k", value=b"v", partition=1, offset=42)
        route_to_dlq(mp, "dlq-topic", rec, ValueError("boom"))

        assert len(mp.records) == 1
        ts = mp.records[0].headers["dlq_timestamp"]
        if isinstance(ts, bytes):
            ts = ts.decode("utf-8")
        # Must parse as a valid ISO timestamp.
        parsed = datetime.fromisoformat(ts)
        assert parsed.tzinfo is not None  # Must be timezone-aware.

    def test_timestamp_not_hardcoded_zeros(self):
        mp = MockProducer()
        rec = Record(topic="src", key=b"k", value=b"v")
        route_to_dlq(mp, "dlq-topic", rec, RuntimeError("fail"))

        ts = mp.records[0].headers["dlq_timestamp"]
        if isinstance(ts, bytes):
            ts = ts.decode("utf-8")
        # Old bug: always ended with ".000Z". New format uses isoformat
        # which produces microseconds (e.g., "+00:00" suffix).
        assert ".000Z" not in ts or "+" in ts



"""Tests for neonlink.record."""

from datetime import datetime, timezone

from neonlink.record import (
    HEADER_CORRELATION_ID,
    HEADER_MESSAGE_ID,
    HEADER_MESSAGE_TYPE,
    HEADER_RETRY_COUNT,
    HEADER_TRACEPARENT,
    HEADER_TRACESTATE,
    Record,
)


class TestRecord:
    def test_get_header_empty(self):
        rec = Record(topic="t")
        assert rec.get_header("missing") == ""

    def test_set_and_get_header(self):
        rec = Record(topic="t")
        rec.set_header(HEADER_MESSAGE_TYPE, "MESSAGE_TYPE_TEST")
        assert rec.message_type == "MESSAGE_TYPE_TEST"

    def test_properties(self):
        rec = Record(
            topic="test-topic",
            key=b"key-1",
            value=b"value-1",
            headers={
                HEADER_MESSAGE_TYPE: "MESSAGE_TYPE_ETL_COMPLETION",
                HEADER_CORRELATION_ID: "corr-123",
                HEADER_MESSAGE_ID: "msg-456",
            },
        )
        assert rec.message_type == "MESSAGE_TYPE_ETL_COMPLETION"
        assert rec.correlation_id == "corr-123"
        assert rec.message_id == "msg-456"

    def test_timestamp_default_none(self):
        rec = Record(topic="t")
        assert rec.timestamp is None

    def test_timestamp_set(self):
        ts = datetime(2026, 3, 1, 12, 0, 0, tzinfo=timezone.utc)
        rec = Record(topic="t", timestamp=ts)
        assert rec.timestamp == ts
        assert rec.timestamp.tzinfo == timezone.utc


class TestPoisonAndRetry:
    def test_poison_pill(self):
        from neonlink.poison import is_poison_pill, get_retry_count, increment_retry_count

        rec = Record(topic="t")
        assert not is_poison_pill(rec, 5)
        assert get_retry_count(rec) == 0

        rec.set_header(HEADER_RETRY_COUNT, "3")
        assert not is_poison_pill(rec, 5)
        assert get_retry_count(rec) == 3

        rec.set_header(HEADER_RETRY_COUNT, "4")
        count = increment_retry_count(rec)
        assert count == 5
        assert is_poison_pill(rec, 5)
        assert get_retry_count(rec) == 5

    def test_invalid_retry_count(self):
        from neonlink.poison import get_retry_count

        rec = Record(topic="t", headers={HEADER_RETRY_COUNT: "invalid"})
        assert get_retry_count(rec) == 0


class TestTracing:
    def test_inject_extract_noop_without_otel(self):
        from neonlink.tracing import inject_trace_context, extract_trace_context

        rec = Record(topic="t")
        inject_trace_context(rec)  # Should not raise
        ctx = extract_trace_context(rec)
        # May return None if OTel not installed, or a context if it is.

    def test_manual_headers(self):
        rec = Record(topic="t")
        rec.set_header(HEADER_TRACEPARENT, "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01")
        rec.set_header(HEADER_TRACESTATE, "rojo=00f067aa0ba902b7")
        assert rec.get_header(HEADER_TRACEPARENT) != ""
        assert rec.get_header(HEADER_TRACESTATE) != ""



"""Tests for TopicRouter."""

import pytest

from neonlink.errors import PermanentError, is_permanent
from neonlink.record import Record
from neonlink.router import TopicRouter


class _StubHandler:
    """Stub handler that records calls."""

    def __init__(self):
        self.calls: list[Record] = []

    def handle_message(self, record: Record) -> None:
        self.calls.append(record)


def _make_record(topic: str, value: bytes = b"") -> Record:
    return Record(topic=topic, value=value)


class TestTopicRouterDispatch:
    def test_dispatches_to_correct_handler(self):
        handler_a = _StubHandler()
        handler_b = _StubHandler()

        router = TopicRouter()
        router.handle("topic-a", handler_a)
        router.handle("topic-b", handler_b)

        router.handle_message(_make_record("topic-a", b"hello"))
        router.handle_message(_make_record("topic-b", b"world"))

        assert len(handler_a.calls) == 1
        assert handler_a.calls[0].value == b"hello"
        assert len(handler_b.calls) == 1
        assert handler_b.calls[0].value == b"world"

    def test_handle_func_dispatches(self):
        called_with: list[Record] = []

        router = TopicRouter()
        router.handle_func("topic-a", lambda rec: called_with.append(rec))

        router.handle_message(_make_record("topic-a", b"payload"))

        assert len(called_with) == 1
        assert called_with[0].value == b"payload"


class TestTopicRouterUnregistered:
    def test_raises_permanent_error_for_unknown_topic(self):
        router = TopicRouter()
        router.handle("topic-a", _StubHandler())

        with pytest.raises(PermanentError):
            router.handle_message(_make_record("unknown"))

    def test_error_is_classified_as_permanent(self):
        router = TopicRouter()
        router.handle("topic-a", _StubHandler())

        try:
            router.handle_message(_make_record("unknown"))
        except PermanentError as exc:
            assert is_permanent(exc)


class TestTopicRouterErrorPropagation:
    def test_handler_error_propagates(self):
        class _FailHandler:
            def handle_message(self, record):
                raise RuntimeError("boom")

        router = TopicRouter()
        router.handle("topic-a", _FailHandler())

        with pytest.raises(RuntimeError, match="boom"):
            router.handle_message(_make_record("topic-a"))


class TestTopicRouterTopics:
    def test_returns_sorted_topics(self):
        router = TopicRouter()
        router.handle("zeta", _StubHandler())
        router.handle("alpha", _StubHandler())
        router.handle("mu", _StubHandler())

        assert router.topics() == ["alpha", "mu", "zeta"]

    def test_empty_router_returns_empty_list(self):
        router = TopicRouter()
        assert router.topics() == []


class TestTopicRouterValidation:
    def test_raises_on_empty_topic(self):
        router = TopicRouter()
        with pytest.raises(ValueError, match="empty topic"):
            router.handle("", _StubHandler())

    def test_raises_on_none_handler(self):
        router = TopicRouter()
        with pytest.raises(ValueError, match="None handler"):
            router.handle("topic-a", None)

    def test_raises_on_none_func(self):
        router = TopicRouter()
        with pytest.raises(ValueError, match="None fn"):
            router.handle_func("topic-a", None)

    def test_raises_on_duplicate_topic(self):
        router = TopicRouter()
        router.handle("topic-a", _StubHandler())
        with pytest.raises(ValueError, match="called twice"):
            router.handle("topic-a", _StubHandler())

    def test_raises_on_invalid_handler_type(self):
        router = TopicRouter()
        with pytest.raises(TypeError, match="must implement handle_message"):
            router.handle("topic-a", "not a handler")



"""Tests for neonlink.config."""

import os
import pytest

from neonlink.config import Config, new_config_from_lookup, new_config_from_map


class TestDefaultConfig:
    def test_defaults(self):
        cfg = Config()
        assert cfg.brokers == ["localhost:9092"]
        assert cfg.max_poll_records == 10
        assert cfg.session_timeout_sec == 30
        assert cfg.auto_offset_reset == "earliest"
        assert cfg.required_acks == "all"
        assert cfg.transactional_id == ""
        assert cfg.cb_max_failures == 5
        assert cfg.retry_max_attempts == 3
        assert cfg.retry_initial_wait_ms == 500
        assert cfg.retry_max_wait_ms == 30000
        assert cfg.dlq_topic == "errors-dlq"
        assert cfg.poison_pill_threshold == 5


class TestNewConfigFromLookup:
    def test_loads_from_lookup(self):
        values = {
            "NEONLINK_BROKERS": "broker1:9092,broker2:9092",
            "NEONLINK_TLS_ENABLED": "true",
            "NEONLINK_CONSUMER_GROUP": "test-group",
            "NEONLINK_MAX_POLL_RECORDS": "50",
            "NEONLINK_SESSION_TIMEOUT_SEC": "45",
            "NEONLINK_AUTO_OFFSET_RESET": "latest",
            "NEONLINK_REQUIRED_ACKS": "all",
            "NEONLINK_TRANSACTIONAL_ID": "txn-test-0",
            "NEONLINK_CB_MAX_FAILURES": "10",
            "NEONLINK_RETRY_MAX_ATTEMPTS": "5",
            "NEONLINK_RETRY_INITIAL_WAIT_MS": "1000",
            "NEONLINK_RETRY_MAX_WAIT_MS": "60000",
            "NEONLINK_DLQ_TOPIC": "custom-dlq",
            "NEONLINK_POISON_THRESHOLD": "3",
            "NEONLINK_CLIENT_ID": "test-client",
        }
        cfg = new_config_from_lookup(lambda k: (values.get(k, ""), k in values))

        assert cfg.brokers == ["broker1:9092", "broker2:9092"]
        assert cfg.tls_enabled is True
        assert cfg.consumer_group == "test-group"
        assert cfg.max_poll_records == 50
        assert cfg.session_timeout_sec == 45
        assert cfg.auto_offset_reset == "latest"
        assert cfg.transactional_id == "txn-test-0"
        assert cfg.cb_max_failures == 10
        assert cfg.retry_max_attempts == 5
        assert cfg.retry_initial_wait_ms == 1000
        assert cfg.retry_max_wait_ms == 60000
        assert cfg.dlq_topic == "custom-dlq"
        assert cfg.poison_pill_threshold == 3
        assert cfg.client_id == "test-client"

    def test_env_backed_parity(self, monkeypatch):
        monkeypatch.setenv("NEONLINK_BROKERS", "env-broker:9092")
        monkeypatch.setenv("NEONLINK_CONSUMER_GROUP", "env-group")

        cfg = new_config_from_lookup(
            lambda k: (os.environ.get(k, ""), k in os.environ)
        )
        assert cfg.brokers == ["env-broker:9092"]
        assert cfg.consumer_group == "env-group"


class TestNewConfigFromMap:
    def test_map_constructor(self):
        cfg = new_config_from_map({
            "NEONLINK_BROKERS": "b1:9092",
            "NEONLINK_CONSUMER_GROUP": "map-group",
        })
        assert cfg.brokers == ["b1:9092"]
        assert cfg.consumer_group == "map-group"

    def test_absent_keys_retain_defaults(self):
        cfg = new_config_from_map({})
        assert cfg.brokers == ["localhost:9092"]
        assert cfg.max_poll_records == 10
        assert cfg.dlq_topic == "errors-dlq"
        assert cfg.required_acks == "all"

    def test_new_fields(self):
        cfg = new_config_from_map({
            "NEONLINK_LINGER_MS": "10",
            "NEONLINK_BATCH_SIZE": "32768",
            "NEONLINK_MAX_MESSAGE_BYTES": "2097152",
            "NEONLINK_MAX_BUFFERED_RECORDS": "500",
        })
        assert cfg.linger_ms == 10
        assert cfg.batch_size == 32768
        assert cfg.max_message_bytes == 2097152
        assert cfg.max_buffered_records == 500


class TestLookupParsingBehaviors:
    def test_invalid_bool_falls_back(self):
        cfg = new_config_from_map({"NEONLINK_TLS_ENABLED": "invalid"})
        assert cfg.tls_enabled is False

    def test_invalid_int_falls_back(self):
        cfg = new_config_from_map({"NEONLINK_MAX_POLL_RECORDS": "not-a-number"})
        assert cfg.max_poll_records == 10

    def test_empty_string_treated_as_absent(self):
        cfg = new_config_from_map({"NEONLINK_CLIENT_ID": ""})
        assert cfg.client_id == ""

class TestConfigValidate:
    def test_valid_default(self):
        Config().validate()

    def test_no_brokers(self):
        with pytest.raises(ValueError, match="at least one broker"):
            Config(brokers=[]).validate()

    def test_empty_broker(self):
        with pytest.raises(ValueError, match="cannot be empty"):
            Config(brokers=[""]).validate()

    def test_tls_without_certs(self):
        with pytest.raises(ValueError, match="TLS enabled"):
            Config(tls_enabled=True).validate()

    def test_sasl_without_credentials(self):
        with pytest.raises(ValueError, match="username/password"):
            Config(sasl_mechanism="PLAIN").validate()

    def test_unsupported_sasl(self):
        with pytest.raises(ValueError, match="unsupported SASL"):
            Config(
                sasl_mechanism="KERBEROS",
                sasl_username="u",
                sasl_password="p",
            ).validate()

    def test_valid_sasl_plain(self):
        Config(
            sasl_mechanism="PLAIN",
            sasl_username="u",
            sasl_password="p",
        ).validate()

    def test_valid_sasl_scram256(self):
        Config(
            sasl_mechanism="SCRAM-SHA-256",
            sasl_username="u",
            sasl_password="p",
        ).validate()

    def test_poison_threshold_zero(self):
        with pytest.raises(ValueError, match="poison pill threshold"):
            Config(poison_pill_threshold=0).validate()


class TestConfluentConfig:
    def test_basic(self):
        cfg = Config(brokers=["b1:9092", "b2:9092"], client_id="test")
        conf = cfg.to_confluent_config()
        assert conf["bootstrap.servers"] == "b1:9092,b2:9092"
        assert conf["client.id"] == "test"

    def test_tls(self):
        cfg = Config(
            tls_enabled=True,
            tls_ca_file="/ca.pem",
            tls_cert_file="/cert.pem",
            tls_key_file="/key.pem",
        )
        conf = cfg.to_confluent_config()
        assert conf["security.protocol"] == "SSL"
        assert conf["ssl.ca.location"] == "/ca.pem"

    def test_sasl_plain(self):
        cfg = Config(
            sasl_mechanism="PLAIN",
            sasl_username="user",
            sasl_password="pass",
        )
        conf = cfg.to_confluent_config()
        assert conf["security.protocol"] == "SASL_PLAINTEXT"
        assert conf["sasl.mechanism"] == "PLAIN"
        assert conf["sasl.username"] == "user"

    def test_sasl_with_tls(self):
        cfg = Config(
            tls_enabled=True,
            tls_ca_file="/ca.pem",
            sasl_mechanism="SCRAM-SHA-512",
            sasl_username="user",
            sasl_password="pass",
        )
        conf = cfg.to_confluent_config()
        assert conf["security.protocol"] == "SASL_SSL"

    def test_producer_config(self):
        cfg = Config()
        conf = cfg.to_producer_config()
        assert conf["acks"] == "all"
        assert conf["retries"] == "3"
        assert conf["retry.backoff.ms"] == "500"
        assert conf["retry.backoff.max.ms"] == "30000"

    def test_producer_config_custom_retry(self):
        cfg = Config(retry_initial_wait_ms=1000, retry_max_wait_ms=60000)
        conf = cfg.to_producer_config()
        assert conf["retry.backoff.ms"] == "1000"
        assert conf["retry.backoff.max.ms"] == "60000"

    def test_consumer_config(self):
        cfg = Config(consumer_group="test-group")
        conf = cfg.to_consumer_config()
        assert conf["group.id"] == "test-group"
        assert conf["enable.auto.commit"] == "false"
        assert conf["auto.offset.reset"] == "earliest"

    def test_transactional_config(self):
        cfg = Config(transactional_id="test-txn-0")
        conf = cfg.to_transactional_config()
        assert conf["acks"] == "all"
        assert conf["transactional.id"] == "test-txn-0"
        assert conf["enable.idempotence"] == "true"

    def test_consumer_config_cooperative_sticky(self):
        cfg = Config(consumer_group="test-group")
        conf = cfg.to_consumer_config()
        assert conf["partition.assignment.strategy"] == "cooperative-sticky"

    def test_producer_config_linger_batch(self):
        cfg = Config(linger_ms=10, batch_size=32768, max_message_bytes=2097152)
        conf = cfg.to_producer_config()
        assert conf["linger.ms"] == "10"
        assert conf["batch.size"] == "32768"
        assert conf["message.max.bytes"] == "2097152"

    def test_transactional_config_requires_id(self):
        cfg = Config()
        with pytest.raises(ValueError, match="transactional_id is required"):
            cfg.to_transactional_config()


class TestConfigNewEnvVars:
    def test_defaults(self):
        cfg = Config()
        assert cfg.linger_ms == 5
        assert cfg.batch_size == 16384
        assert cfg.max_message_bytes == 1048576
        assert cfg.max_buffered_records == 500


class TestPayloadValidation:
    def test_payload_too_large(self):
        from neonlink.errors import PayloadTooLargeError

        with pytest.raises(PayloadTooLargeError):
            raise PayloadTooLargeError("too big")



"""Record type wrapping a Kafka message with typed header accessors."""

from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional


# Header key constants.
HEADER_MESSAGE_TYPE = "message_type"
HEADER_SOURCE_SERVICE = "source_service"
HEADER_TARGET_SERVICE = "target_service"
HEADER_MESSAGE_ID = "message_id"
HEADER_CORRELATION_ID = "correlation_id"
HEADER_IDENTITY_CTX = "identity_context"
HEADER_RETRY_COUNT = "retry_count"
HEADER_TRACEPARENT = "traceparent"
HEADER_TRACESTATE = "tracestate"

Headers = dict[str, str | bytes]


@dataclass
class Record:
    """A message consumed from or produced to a broker topic."""

    topic: str = ""
    key: Optional[bytes] = None
    value: Optional[bytes] = None
    headers: Headers = field(default_factory=dict)
    partition: int = 0
    offset: int = 0
    timestamp: Optional[datetime] = None

    def get_header(self, key: str) -> str:
        """Return header value as string, or empty string if not found.

        Bytes values are decoded as UTF-8 with replacement characters for
        backward compatibility with callers that expect str.
        """
        val = self.headers.get(key)
        if val is None:
            return ""
        if isinstance(val, bytes):
            return val.decode("utf-8", errors="replace")
        return val

    def get_header_bytes(self, key: str) -> bytes | None:
        """Return header value as raw bytes, or None if not found."""
        val = self.headers.get(key)
        if val is None:
            return None
        if isinstance(val, str):
            return val.encode("utf-8")
        return val

    def set_header(self, key: str, value: str) -> None:
        """Set a header key-value pair (str, for backward compatibility)."""
        self.headers[key] = value

    def set_header_bytes(self, key: str, value: bytes) -> None:
        """Set a header key-value pair with raw bytes."""
        self.headers[key] = value

    @property
    def message_type(self) -> str:
        return self.get_header(HEADER_MESSAGE_TYPE)

    @property
    def correlation_id(self) -> str:
        return self.get_header(HEADER_CORRELATION_ID)

    @property
    def message_id(self) -> str:
        return self.get_header(HEADER_MESSAGE_ID)


# ---------------------------------------------------------------------------
# Wire-size estimation (shared by producer and transactional producer)
# ---------------------------------------------------------------------------

HEADER_FRAMING_OVERHEAD = 6
"""Per-header framing cost in the Kafka wire protocol."""

PROTOCOL_OVERHEAD = 128
"""Conservative estimate of Kafka record protocol overhead."""


def encoded_record_size(
    key: Optional[bytes],
    value: Optional[bytes],
    headers: Optional[dict[str, str | bytes]],
) -> int:
    """Estimate the total wire size of a Kafka record."""
    size = (len(key) if key else 0) + (len(value) if value else 0) + PROTOCOL_OVERHEAD
    if headers:
        for k, v in headers.items():
            size += len(k) + (len(v) if v else 0) + HEADER_FRAMING_OVERHEAD
    return size


def headers_to_list(headers: dict[str, str | bytes]) -> list[tuple[str, bytes]]:
    """Convert header dict to confluent-kafka format."""
    return [(k, v if isinstance(v, bytes) else v.encode("utf-8")) for k, v in headers.items()]



"""Tests for neonlink.health."""

from neonlink.config import Config


class TestHealthCheckerConfig:
    def test_health_client_id_suffix(self):
        cfg = Config(client_id="test-service")
        conf = cfg.to_confluent_config()
        assert conf["client.id"] == "test-service"
        # HealthChecker appends "-health" internally; verify config is valid.

    def test_health_with_tls_sasl(self):
        cfg = Config(
            tls_enabled=True,
            tls_ca_file="/ca.pem",
            sasl_mechanism="SCRAM-SHA-512",
            sasl_username="admin",
            sasl_password="secret",
        )
        conf = cfg.to_confluent_config()
        assert conf["security.protocol"] == "SASL_SSL"
        assert conf["ssl.ca.location"] == "/ca.pem"
        assert conf["sasl.mechanism"] == "SCRAM-SHA-512"



"""Tests for neonlink.transaction."""

import pytest

from neonlink.config import Config
from neonlink.record import Record
from neonlink.transaction import TransactionalProducer


class TestTransactionalProducerValidation:
    def test_requires_transactional_id(self):
        cfg = Config()
        assert cfg.transactional_id == ""

    def test_transactional_config_requires_id(self):
        cfg = Config()
        with pytest.raises(ValueError, match="transactional_id is required"):
            cfg.to_transactional_config()

    def test_transactional_config_forces_acks_all(self):
        cfg = Config(transactional_id="test-txn-0")
        conf = cfg.to_transactional_config()
        assert conf["acks"] == "all"
        assert conf["transactional.id"] == "test-txn-0"
        assert conf["enable.idempotence"] == "true"

    def test_transactional_config_with_tls_sasl(self):
        cfg = Config(
            transactional_id="test-txn-1",
            tls_enabled=True,
            tls_ca_file="/ca.pem",
            sasl_mechanism="SCRAM-SHA-256",
            sasl_username="user",
            sasl_password="pass",
        )
        conf = cfg.to_transactional_config()
        assert conf["security.protocol"] == "SASL_SSL"
        assert conf["transactional.id"] == "test-txn-1"
        assert conf["sasl.mechanism"] == "SCRAM-SHA-256"


class TestTransactionalIDFromLookup:
    def test_loads_from_map(self):
        from neonlink.config import new_config_from_map

        cfg = new_config_from_map({"NEONLINK_TRANSACTIONAL_ID": "findata-coordinator-0"})
        assert cfg.transactional_id == "findata-coordinator-0"

    def test_default_empty(self):
        cfg = Config()
        assert cfg.transactional_id == ""


class TestTransactionalProducerUniversalMethods:
    def test_publish_delegates_to_publish_atomic(self):
        p = object.__new__(TransactionalProducer)
        called = {}

        def _fake_publish_atomic(records, *, cancel_event=None, timeout_sec=30.0):
            called["records"] = records
            called["cancel_event"] = cancel_event
            called["timeout_sec"] = timeout_sec

        p.publish_atomic = _fake_publish_atomic

        p.publish(
            "topic-a",
            b"k",
            b"v",
            {"h": "v"},
            timeout_sec=12.0,
        )

        assert len(called["records"]) == 1
        rec = called["records"][0]
        assert rec.topic == "topic-a"
        assert rec.key == b"k"
        assert rec.value == b"v"
        assert rec.headers == {"h": "v"}
        assert called["timeout_sec"] == 12.0

    def test_publish_record_delegates_to_publish_atomic(self):
        p = object.__new__(TransactionalProducer)
        called = {}

        def _fake_publish_atomic(records, *, cancel_event=None, timeout_sec=30.0):
            called["records"] = records
            called["cancel_event"] = cancel_event
            called["timeout_sec"] = timeout_sec

        p.publish_atomic = _fake_publish_atomic

        rec = Record(topic="topic-b", key=b"k2", value=b"v2")
        p.publish_record(rec, timeout_sec=9.0)

        assert called["records"] == [rec]
        assert called["timeout_sec"] == 9.0

    def test_publish_batch_delegates_to_publish_atomic(self):
        p = object.__new__(TransactionalProducer)
        called = {}

        def _fake_publish_atomic(records, *, cancel_event=None, timeout_sec=30.0):
            called["records"] = records
            called["cancel_event"] = cancel_event
            called["timeout_sec"] = timeout_sec

        p.publish_atomic = _fake_publish_atomic

        recs = [Record(topic="topic-1"), Record(topic="topic-2")]
        p.publish_batch(recs, timeout_sec=5.0)

        assert called["records"] == recs
        assert called["timeout_sec"] == 5.0



"""Post-remediation consumer hardening tests."""

from __future__ import annotations

import threading

from neonlink.circuit_breaker import CircuitBreaker, CircuitBreakerState
from neonlink.config import Config
from neonlink.consumer import Consumer


class _FakeMessage:
    def __init__(
        self,
        topic: str = "topic-a",
        partition: int = 0,
        offset: int = 1,
        err=None,
    ) -> None:
        self._topic = topic
        self._partition = partition
        self._offset = offset
        self._err = err

    def error(self):
        return self._err

    def topic(self):
        return self._topic

    def partition(self):
        return self._partition

    def offset(self):
        return self._offset

    def headers(self):
        return []

    def key(self):
        return b"k"

    def value(self):
        return b"v"

    def timestamp(self):
        return (0, 0)


class _FakePollError:
    def __init__(self, code: int) -> None:
        self._code = code

    def code(self):
        return self._code


class _FakeCommitConsumer:
    def __init__(self) -> None:
        self.commit_calls = 0

    def commit(self, message=None, asynchronous=False):
        self.commit_calls += 1


def test_commit_fenced_for_revoked_partition():
    c = object.__new__(Consumer)
    c._revoked_partitions = {"topic-a:1"}
    c._revoked_lock = threading.Lock()
    c._consumer = _FakeCommitConsumer()

    ok = Consumer._commit(c, _FakeMessage(topic="topic-a", partition=1, offset=44))
    assert ok is False
    assert c._consumer.commit_calls == 0


def test_subscribe_with_dlq_forwards_cancel_event():
    c = object.__new__(Consumer)
    called = {}

    def _fake_subscribe(handler, *, cancel_event=None):
        called["handler"] = handler
        called["cancel_event"] = cancel_event

    c.subscribe = _fake_subscribe

    handler = object()
    dlq = object()
    ev = threading.Event()
    Consumer.subscribe_with_dlq(c, handler, dlq, cancel_event=ev)

    assert c._dlq_producer is dlq
    assert called["handler"] is handler
    assert called["cancel_event"] is ev


def test_poll_message_error_trips_circuit_breaker(monkeypatch):
    import neonlink.consumer as consumer_mod

    class _DummyKafkaException(Exception):
        pass

    class _DummyKafkaError:
        _PARTITION_EOF = -191

    class _FakePollConsumer:
        def poll(self, timeout=1.0):
            return _FakeMessage(err=_FakePollError(code=123))

    monkeypatch.setattr(consumer_mod, "KafkaException", _DummyKafkaException)
    monkeypatch.setattr(consumer_mod, "KafkaError", _DummyKafkaError)

    c = object.__new__(Consumer)
    c._consumer = _FakePollConsumer()
    c._cb = CircuitBreaker("poll", max_failures=1, reset_timeout_sec=60)
    c._last_poll_error = False

    msg = Consumer._poll_one(c)
    assert msg is None
    assert c._last_poll_error is True
    assert c._cb.state == CircuitBreakerState.OPEN


def test_handler_with_commit_controls_ack(monkeypatch):
    import neonlink.consumer as consumer_mod

    class _FakeConfluentConsumer:
        def __init__(self, _cfg):
            self.commit_calls = 0
            self.poll_calls = 0

        def subscribe(self, _topics, on_revoke=None, on_assign=None):
            self._on_revoke = on_revoke
            self._on_assign = on_assign

        def poll(self, timeout=1.0):
            if self.poll_calls == 0:
                self.poll_calls += 1
                return _FakeMessage(topic="topic-a", partition=0, offset=10)
            return None

        def commit(self, message=None, asynchronous=False):
            self.commit_calls += 1

        def assignment(self):
            return []

        def pause(self, _parts):
            return None

        def resume(self, _parts):
            return None

        def get_watermark_offsets(self, _tp, cached=True):
            return (0, 11)

        def close(self):
            return None

    monkeypatch.setattr(consumer_mod, "ConfluentConsumer", _FakeConfluentConsumer)

    cfg = Config(consumer_group="group-a")
    c = Consumer(cfg, topics=["topic-a"])
    stop_event = threading.Event()

    class _Handler:
        def __init__(self):
            self.called_manual = False
            self.called_auto = False

        def handle_message(self, _record):
            self.called_auto = True

        def handle_message_with_commit(self, _record, commit):
            self.called_manual = True
            assert commit() is True
            stop_event.set()

    h = _Handler()
    c.subscribe(h, cancel_event=stop_event)

    assert h.called_manual is True
    assert h.called_auto is False
    assert c._consumer.commit_calls == 1



"""Metrics interface for SDK observability.

Implementations bridge to Prometheus, OTel, or custom backends.
The SDK ships with NoopMetrics (zero overhead) to avoid forcing
a metrics dependency.
"""

from __future__ import annotations

from typing import Protocol


class Metrics(Protocol):
    """Protocol defining the metrics contract for the NeonLink SDK."""

    def publish_success(self, topic: str, latency_sec: float) -> None: ...
    def publish_error(self, topic: str, err: Exception) -> None: ...
    def consume_success(self, topic: str, partition: int, latency_sec: float) -> None: ...
    def consume_error(self, topic: str, partition: int, err: Exception) -> None: ...
    def poison_pill_detected(self, topic: str, partition: int, offset: int) -> None: ...
    def poison_pill_blocked(self, topic: str, partition: int, offset: int) -> None: ...
    def consumer_lag(self, topic: str, partition: int, lag: int) -> None:
        """Report per-fetch processing lag (high watermark - current offset).

        This is NOT committed-offset group lag. For true consumer group lag,
        use Kafka admin APIs externally.
        """
        ...
    def circuit_breaker_state_change(self, name: str, from_state: str, to_state: str) -> None: ...
    def dlq_routed(self, original_topic: str, err: Exception) -> None: ...


class NoopMetrics:
    """Default metrics implementation with zero overhead. All methods are no-ops."""

    def publish_success(self, topic: str, latency_sec: float) -> None:
        pass

    def publish_error(self, topic: str, err: Exception) -> None:
        pass

    def consume_success(self, topic: str, partition: int, latency_sec: float) -> None:
        pass

    def consume_error(self, topic: str, partition: int, err: Exception) -> None:
        pass

    def poison_pill_detected(self, topic: str, partition: int, offset: int) -> None:
        pass

    def poison_pill_blocked(self, topic: str, partition: int, offset: int) -> None:
        pass

    def consumer_lag(self, topic: str, partition: int, lag: int) -> None:
        pass

    def circuit_breaker_state_change(self, name: str, from_state: str, to_state: str) -> None:
        pass

    def dlq_routed(self, original_topic: str, err: Exception) -> None:
        pass



"""Tests for neonlink.errors."""

from neonlink.errors import (
    CircuitBreakerOpenError,
    ErrorClass,
    PayloadTooLargeError,
    PermanentError,
    classify,
    is_permanent,
    is_transient,
)


class TestClassify:
    def test_none(self):
        # classify(None) should return UNKNOWN
        assert classify(None) == ErrorClass.UNKNOWN

    def test_permanent_error(self):
        err = PermanentError(ValueError("bad input"))
        assert classify(err) == ErrorClass.PERMANENT
        assert is_permanent(err)
        assert not is_transient(err)

    def test_permanent_error_string(self):
        err = PermanentError("bad schema")
        assert classify(err) == ErrorClass.PERMANENT
        assert "permanent: bad schema" in str(err)

    def test_permanent_error_none(self):
        err = PermanentError()
        assert str(err) == "permanent error"

    def test_payload_too_large(self):
        err = PayloadTooLargeError("1MB exceeded")
        assert classify(err) == ErrorClass.PERMANENT

    def test_permanent_keywords(self):
        keywords = [
            "validation failed", "invalid input", "schema mismatch",
            "constraint violation", "auth denied", "forbidden access",
            "permission denied", "malformed request", "syntax error",
        ]
        for kw in keywords:
            err = Exception(kw)
            assert classify(err) == ErrorClass.PERMANENT, f"expected PERMANENT for {kw!r}"

    def test_transient_keywords(self):
        keywords = [
            "connection refused", "network unreachable", "timeout exceeded",
            "temporary failure", "rate limited", "overloaded",
            "broker busy", "service unavailable", "retry later",
        ]
        for kw in keywords:
            err = Exception(kw)
            assert classify(err) == ErrorClass.TRANSIENT, f"expected TRANSIENT for {kw!r}"

    def test_unknown(self):
        err = Exception("something happened")
        assert classify(err) == ErrorClass.UNKNOWN

    def test_circuit_breaker_open_error(self):
        err = CircuitBreakerOpenError("open")
        # Not classified as permanent or transient by keywords
        assert classify(err) is not None


class TestErrorClassString:
    def test_values(self):
        assert ErrorClass.TRANSIENT.value == "TRANSIENT"
        assert ErrorClass.PERMANENT.value == "PERMANENT"
        assert ErrorClass.UNKNOWN.value == "UNKNOWN"



"""NeonLink Kafka client library for Python services.

Thin wrapper around confluent-kafka-python providing:
- Producer with circuit breaker, error classification, and payload validation
- TransactionalProducer for atomic multi-topic writes
- Consumer with manual offset commit, failure tracking, and panic recovery
- DLQ routing with enriched forensics headers
- HealthChecker for broker readiness probes
- OpenTelemetry W3C trace context propagation via Kafka headers
- Error taxonomy (PermanentError, Classify, IsTransient, IsPermanent)
- Metrics interface with NoopMetrics default
- MockProducer and MockConsumer for testing
"""

from neonlink.config import Config, new_config_from_lookup, new_config_from_map
from neonlink.producer import Producer
from neonlink.transaction import TransactionalProducer
from neonlink.consumer import Consumer, MessageHandler, MessageHandlerWithCommit
from neonlink.health import HealthChecker
from neonlink.record import Record, Headers, encoded_record_size
from neonlink.dlq import route_to_dlq
from neonlink.poison import is_poison_pill, get_retry_count, increment_retry_count
from neonlink.tracing import inject_trace_context, extract_trace_context
from neonlink.errors import (
    ErrorClass,
    PermanentError,
    CircuitBreakerOpenError,
    PayloadTooLargeError,
    classify,
    is_transient,
    is_permanent,
)
from neonlink.metrics import Metrics, NoopMetrics
from neonlink.mock import MockProducer, MockConsumer
from neonlink.router import TopicRouter

__all__ = [
    "Config",
    "new_config_from_lookup",
    "new_config_from_map",
    "Producer",
    "TransactionalProducer",
    "Consumer",
    "MessageHandler",
    "MessageHandlerWithCommit",
    "HealthChecker",
    "Record",
    "Headers",
    "encoded_record_size",
    "route_to_dlq",
    "is_poison_pill",
    "get_retry_count",
    "increment_retry_count",
    "inject_trace_context",
    "extract_trace_context",
    "ErrorClass",
    "PermanentError",
    "CircuitBreakerOpenError",
    "PayloadTooLargeError",
    "classify",
    "is_transient",
    "is_permanent",
    "Metrics",
    "NoopMetrics",
    "MockProducer",
    "MockConsumer",
    "TopicRouter",
]



"""OpenTelemetry W3C trace context propagation via Kafka headers."""

from __future__ import annotations

from typing import Optional

from neonlink.record import HEADER_TRACEPARENT, HEADER_TRACESTATE, Record

try:
    from opentelemetry import context as otel_context
    from opentelemetry import trace
    from opentelemetry.trace import StatusCode
    from opentelemetry.trace.propagation import TraceContextTextMapPropagator

    _HAS_OTEL = True
    _propagator = TraceContextTextMapPropagator()
    _consumer_tracer = trace.get_tracer("neonlink.kafka.consumer")
    _producer_tracer = trace.get_tracer("neonlink.kafka.producer")
except ImportError:
    _HAS_OTEL = False
    _propagator = None
    _consumer_tracer = None
    _producer_tracer = None


class _RecordCarrier:
    """Adapter between Record headers and OTel TextMapPropagator."""

    def __init__(self, record: Record) -> None:
        self._record = record

    def get(self, key: str) -> Optional[str]:
        val = self._record.headers.get(key)
        if val is None:
            return None
        if isinstance(val, bytes):
            return val.decode("utf-8", errors="replace")
        return val or None

    def set(self, key: str, value: str) -> None:
        self._record.headers[key] = value.encode("utf-8")

    def keys(self) -> list[str]:
        return list(self._record.headers.keys())


def inject_trace_context(record: Record) -> None:
    """Inject the current span's W3C trace context into record headers.

    Call this before producing a record. If OpenTelemetry is not installed,
    this is a no-op.
    """
    if not _HAS_OTEL:
        return
    _propagator.inject(carrier=_RecordCarrier(record))


def extract_trace_context(record: Record) -> Optional[object]:
    """Extract W3C trace context from record headers.

    Returns an OTel Context object with the remote span context attached.
    If OpenTelemetry is not installed, returns None.
    """
    if not _HAS_OTEL:
        return None
    return _propagator.extract(carrier=_RecordCarrier(record))


class _NoOpSpan:
    """Minimal no-op span for when OTel is not installed."""

    def set_status(self, *args, **kwargs):
        pass

    def record_error(self, *args, **kwargs):
        pass

    def end(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


_NOOP_SPAN = _NoOpSpan()


def start_consumer_span(record: Record):
    """Start a consumer span for a consumed record.

    Returns (span,). Callers must call span.end() when processing is complete.
    If OTel is not installed, returns a no-op span.
    """
    if not _HAS_OTEL:
        return _NOOP_SPAN
    return _consumer_tracer.start_span(
        f"{record.topic} receive",
        kind=trace.SpanKind.CONSUMER,
        attributes={
            "messaging.system": "kafka",
            "messaging.destination": record.topic,
            "messaging.kafka.offset": record.offset,
            "messaging.kafka.partition": record.partition,
        },
    )


def start_producer_span(topic: str):
    """Start a producer span for a produced record.

    Returns (span,). Callers must call span.end() when publishing is complete.
    If OTel is not installed, returns a no-op span.
    """
    if not _HAS_OTEL:
        return _NOOP_SPAN
    return _producer_tracer.start_span(
        f"{topic} send",
        kind=trace.SpanKind.PRODUCER,
        attributes={
            "messaging.system": "kafka",
            "messaging.destination": topic,
        },
    )


def end_span_with_error(span, err: Optional[Exception] = None) -> None:
    """End a span, optionally recording an error."""
    if err is not None and _HAS_OTEL:
        span.record_error(err)
        span.set_status(StatusCode.ERROR, str(err))
    span.end()



"""Broker health checks for readiness probes."""

from __future__ import annotations

from confluent_kafka.admin import AdminClient

from neonlink.config import Config


class HealthChecker:
    """Provides broker health checks for readiness probes.

    Uses the confluent-kafka AdminClient to query cluster metadata.
    """

    def __init__(self, cfg: Config) -> None:
        conf = cfg.to_confluent_config()
        if cfg.client_id:
            conf["client.id"] = cfg.client_id + "-health"
        self._admin = AdminClient(conf)

    def check(self, timeout_sec: float = 5.0) -> None:
        """Verify broker connectivity by requesting cluster metadata.

        Raises:
            RuntimeError: if the cluster is unreachable or returns no brokers.
        """
        metadata = self._admin.list_topics(timeout=timeout_sec)
        if not metadata.brokers:
            raise RuntimeError("neonlink: health check failed: no brokers returned")

    def check_topic_exists(self, topic: str, timeout_sec: float = 5.0) -> None:
        """Verify that a specific topic exists in the cluster.

        Raises:
            RuntimeError: if the topic does not exist or the cluster is unreachable.
        """
        metadata = self._admin.list_topics(timeout=timeout_sec)
        if topic not in metadata.topics:
            raise RuntimeError(f"neonlink: topic {topic!r} does not exist")



"""Error classification for circuit breaker and retry decisions."""

from __future__ import annotations

from enum import Enum


class ErrorClass(Enum):
    """Classifies an error for circuit breaker and retry decisions."""

    TRANSIENT = "TRANSIENT"
    PERMANENT = "PERMANENT"
    UNKNOWN = "UNKNOWN"


class PermanentError(Exception):
    """Wraps an exception to mark it as non-retriable.

    The circuit breaker will pass these through without incrementing the
    failure counter.
    """

    def __init__(self, cause: Exception | str | None = None) -> None:
        self.cause = cause
        if isinstance(cause, Exception):
            super().__init__(f"permanent: {cause}")
        elif isinstance(cause, str):
            super().__init__(f"permanent: {cause}")
        else:
            super().__init__("permanent error")


class CircuitBreakerOpenError(Exception):
    """Raised when the circuit breaker is open and rejecting all calls."""

    pass


class PayloadTooLargeError(Exception):
    """Raised when a message payload exceeds the configured limit."""

    pass


class MalformedProtobufError(PermanentError):
    """Raised when protobuf deserialization fails or panics.

    Inherits from PermanentError so it does not trip the circuit breaker.
    """

    def __init__(self, cause: Exception | str | None = None) -> None:
        super().__init__(cause or "neonlink: malformed protobuf")


# Keywords that indicate a permanent, non-retriable error.
_PERMANENT_KEYWORDS = (
    "validation", "invalid", "permanent", "schema", "constraint",
    "auth", "forbidden", "permission", "malformed", "syntax",
)

# Keywords that indicate a transient, retriable error.
_TRANSIENT_KEYWORDS = (
    "timeout", "connection", "network", "temporary", "rate",
    "limit", "overload", "busy", "unavailable", "retry",
)


# Kafka error codes that are definitively non-retriable.
_PERMANENT_KAFKA_CODES = frozenset({
    10,  # MSG_SIZE_TOO_LARGE
    2,   # INVALID_MSG / CORRUPT_MESSAGE
    29,  # TOPIC_AUTHORIZATION_FAILED
    30,  # GROUP_AUTHORIZATION_FAILED
    31,  # CLUSTER_AUTHORIZATION_FAILED
    17,  # INVALID_TOPIC_EXCEPTION (UNKNOWN_TOPIC_OR_PART)
    87,  # INVALID_RECORD
})


def _classify_kafka_error(err: Exception) -> ErrorClass | None:
    """Attempt typed classification of confluent-kafka errors.

    Returns ErrorClass if the error is a KafkaException/KafkaError, None otherwise.
    """
    try:
        from confluent_kafka import KafkaException, KafkaError
    except ImportError:
        return None

    kafka_err = None
    if isinstance(err, KafkaException):
        args = err.args
        if args and isinstance(args[0], KafkaError):
            kafka_err = args[0]
    elif isinstance(err, KafkaError):
        kafka_err = err

    if kafka_err is None:
        return None

    code = kafka_err.code()
    if code in _PERMANENT_KAFKA_CODES:
        return ErrorClass.PERMANENT
    if kafka_err.retriable():
        return ErrorClass.TRANSIENT
    if kafka_err.fatal():
        return ErrorClass.PERMANENT

    # Non-retriable, non-fatal Kafka errors default to permanent.
    return ErrorClass.PERMANENT


def classify(err: Exception) -> ErrorClass:
    """Determine whether an error is transient or permanent.

    Uses type checks first, then typed Kafka error classification,
    then falls back to keyword matching on the error message.
    """
    if err is None:
        return ErrorClass.UNKNOWN

    # Type assertion: PermanentError is always permanent.
    if isinstance(err, PermanentError):
        return ErrorClass.PERMANENT

    # Sentinel types.
    if isinstance(err, PayloadTooLargeError):
        return ErrorClass.PERMANENT

    # Typed Kafka error classification (before keyword fallback).
    kafka_class = _classify_kafka_error(err)
    if kafka_class is not None:
        return kafka_class

    # Keyword matching on the error message (last resort for non-Kafka errors).
    msg = str(err).lower()

    for kw in _PERMANENT_KEYWORDS:
        if kw in msg:
            return ErrorClass.PERMANENT

    for kw in _TRANSIENT_KEYWORDS:
        if kw in msg:
            return ErrorClass.TRANSIENT

    return ErrorClass.UNKNOWN


def is_transient(err: Exception) -> bool:
    """Return True if the error is classified as transient (retriable)."""
    return classify(err) == ErrorClass.TRANSIENT


def is_permanent(err: Exception) -> bool:
    """Return True if the error is classified as permanent (non-retriable)."""
    return classify(err) == ErrorClass.PERMANENT


def safe_decode(fn):
    """Wrap a protobuf decode callable with exception containment.

    Calls ``fn()`` and catches any exception, re-raising as
    MalformedProtobufError (a PermanentError) to prevent malformed payloads
    from crashing the consumer.

    Usage::

        safe_decode(lambda: job_dispatch.ParseFromString(record.value))
    """
    try:
        fn()
    except MalformedProtobufError:
        raise
    except Exception as exc:
        raise MalformedProtobufError(exc) from exc
